"""Implementation of the vector workloads."""
from __future__ import annotations

import logging
from typing import Any, Union

import geopandas as gpd
import numpy as np
import pyproj
import shapely
import xarray as xr
from geopandas import GeoDataFrame, GeoSeries
from rasterio import gdal_version
from shapely.geometry.base import BaseGeometry

from hydromt import gis_utils, raster

logger = logging.getLogger(__name__)
GDAL_VERSION = gdal_version()


class GeoBase(raster.XGeoBase):

    """Base accessor class for geo data."""

    def __init__(self, xarray_obj):
        """Initialize a new object based on the provided xarray_obj."""
        super(GeoBase, self).__init__(xarray_obj)
        self._geometry = None

    @property
    def _all_names(self):
        """Return names of all variables and coordinates in the dataset/array."""
        # TODO: move to geobase
        names = [n for n in self._obj.coords]
        if isinstance(self._obj, xr.Dataset):
            names = names + [n for n in self._obj.data_vars]
        return names

    def _get_geom_names_types(self, geom_name: str = None) -> tuple[list, list]:
        """Discover coordinates with wkt/geom type data the dataset/array."""
        names, types = [], []
        dvars = self._all_names if geom_name is None else [geom_name]
        for name in dvars:
            if self._obj[name].ndim == 1 and isinstance(
                self._obj[name][0].values.item(), BaseGeometry
            ):
                names.append(name)
                types.append("geom")
            elif self._obj[name].ndim == 1 and isinstance(
                self._obj[name][0].values.item(), str
            ):
                try:
                    shapely.wkt.loads(self._obj[name][0].values.item())
                    names.append(name)
                    types.append("wkt")
                except Exception:
                    pass
        return names, types

    def _discover_xy(self, x_name=None, y_name=None, index_dim=None):
        """Discover xy type geometries in the dataset/array."""
        # infer x dim
        if x_name is None:
            for name in raster.XDIMS:
                if name in self._all_names:
                    dim0 = (
                        index_dim if index_dim is not None else self._obj[name].dims[0]
                    )
                    if self._obj[name].ndim == 1 and self._obj[name].dims[0] == dim0:
                        x_name = name
                        index_dim = dim0
                        break
        # infer y dim
        if y_name is None and x_name is not None:
            for name in raster.YDIMS:
                if name in self._all_names:
                    if self._obj[name].dims[0] == index_dim:
                        y_name = name
                        break
        if (
            x_name is not None
            and y_name is not None
            and self._obj[x_name].ndim == 1
            and self._obj[x_name].dims == self._obj[y_name].dims
        ):
            self.set_attrs(x_name=x_name)
            self.set_attrs(y_name=y_name)
            self.set_attrs(geom_format="xy")
            self.set_attrs(index_dim=index_dim)
        else:
            self.attrs.pop("x_name", None)
            self.attrs.pop("y_name", None)
            self.attrs.pop("geom_format", None)
            self.attrs.pop("index_dim", None)

    def _discover_geom(self, geom_name=None, index_dim=None):
        """Discover geom/wkt type geometries in the dataset/array."""
        # check /infer geom dim
        names, types = self._get_geom_names_types(geom_name=geom_name)
        if geom_name is None:
            if index_dim is not None:
                names = [name for name in names if self._obj[name].dims[0] == index_dim]
            # priority for geometry named variable/coord
            if "geometry" in names:
                geom_name = "geometry"
            # otherwise take first of type geom
            elif len(names) >= 1:
                idx = 0
                if "geom" in types:
                    idx = types.index("geom")
                geom_name = names[idx]
        elif geom_name not in names:
            raise ValueError(f"{geom_name} variable not recognized as geometry.")

        if geom_name is not None:
            self.set_attrs(geom_name=geom_name)
            self.set_attrs(geom_format=types[names.index(geom_name)])
            self.set_attrs(index_dim=self._obj[geom_name].dims[0])
        else:
            self.attrs.pop("geom_name", None)
            self.attrs.pop("geom_format", None)
            self.attrs.pop("index_dim", None)

    def set_spatial_dims(
        self,
        geom_name: str = None,
        x_name: str = None,
        y_name: str = None,
        index_dim: str = None,
        geom_format: str = None,
    ) -> None:
        """Set the spatial and index dimensions of the object.

        Arguments
        ---------
        x_name: str, optional
            The name of the x coordinate.
        y_name: str, optional
            The name of the ycoordinate.
        geom_name: str, optional
            The name of the geometry coordinate.
        index_dim: str optional
            The name of the geometry index dimension
        geom_format: {'xy', 'wkt', 'geom'}
            Geometry format
        """
        if geom_format != "xy":
            self._discover_geom(geom_name=geom_name, index_dim=index_dim)
            self.attrs.pop("x_name", None)
            self.attrs.pop("y_name", None)
        if "geom_name" not in self.attrs:
            self._discover_xy(x_name=x_name, y_name=y_name, index_dim=index_dim)
            self.attrs.pop("geom_name", None)
        if "geom_format" not in self.attrs:
            raise ValueError("No geometry data found.")

    @property
    def geom_format(self) -> str:
        """Name of geometry coordinate; only for 'wkt' and 'geom' formats."""
        if self.get_attrs("geom_format") not in self._obj.dims:
            self.set_spatial_dims()
        if "geom_format" in self.attrs:
            return self.attrs["geom_format"]

    @property
    def geom_name(self) -> str:
        """Name of geometry coordinate; only for 'wkt' and 'geom' formats."""
        if self.get_attrs("geom_name") not in self._obj.dims:
            self.set_spatial_dims()
        if "geom_name" in self.attrs:
            return self.attrs["geom_name"]

    @property
    def geom_type(self) -> str:
        """Return geometry type."""
        geom_types = self.geometry.type.values
        if len(set(geom_types)) > 1:
            # TODO: can we safely assume there is multi geom?
            i = ["MULTI" in g.upper() for g in geom_types].index(True)
            geom_type = geom_types[i]
        else:
            geom_type = geom_types[0]
        del geom_types
        return geom_type

    @property
    def x_name(self):
        """Name of x coordinate; only for point geometries in xy format."""
        if self.get_attrs("x_name") not in self._obj.dims:
            self.set_spatial_dims()
        if "x_name" in self.attrs:
            return self.attrs["x_name"]

    @property
    def y_name(self):
        """Name of y coordinate; only for point geometries in xy format."""
        if self.get_attrs("y_name") not in self._obj.dims:
            self.set_spatial_dims()
        if "y_name" in self.attrs:
            return self.attrs["y_name"]

    @property
    def index_dim(self):
        """Index dimension name."""
        if self.get_attrs("index_dim") not in self._obj.dims:
            self.set_spatial_dims()
        if "index_dim" in self.attrs:
            return self.attrs["index_dim"]

    @property
    def sindex(self):
        """Return the spatial index of the geometry."""
        return self.geometry.sindex

    @property
    def index(self):
        """Return the index values."""
        return self._obj[self.index_dim]

    @property
    def size(self):
        """Return the length of the index array."""
        return self._obj[self.index_dim].size

    @property
    def bounds(self):
        """Return the bounds (xmin, ymin, xmax, ymax) of the object."""
        return self.geometry.total_bounds

    @property
    def geometry(self) -> GeoSeries:
        """Return the geometry of the dataset or array as GeoSeries.

        Returns
        -------
        GeoSeries
            A Series object with shapely geometries
        """
        if self._geometry is not None and self._geometry.index.size == self.size:
            return self._geometry
        # if no geometry is present return None self._geometry
        # rather than raising an error
        try:
            self.set_spatial_dims()
        except ValueError:
            return self._geometry
        gtype = self.geom_format
        if gtype not in ["geom", "xy", "wkt"]:
            raise ValueError("No valid geometry found in object.")
        if gtype == "geom":
            geoms = GeoSeries(
                data=self._obj[self.geom_name].values,
                index=self.index.values,
                crs=self.crs,
            )
        elif gtype == "xy":
            geoms = GeoSeries.from_xy(
                x=self._obj[self.x_name].values,
                y=self._obj[self.y_name].values,
                index=self.index.values,
                crs=self.crs,
            )
        elif gtype == "wkt":
            geoms = GeoSeries.from_wkt(
                data=self._obj[self.geom_name].values,
                index=self.index.values,
                crs=self.crs,
            )
        geoms.index.name = self.index_dim
        self._geometry = geoms
        return geoms

    def update_geometry(
        self,
        geometry: GeoSeries = None,
        geom_format: str = None,
        x_name: str = None,
        y_name: str = None,
        geom_name: str = None,
        replace: bool = True,
    ):
        """Update the geometry in the Dataset/Array with a new geometry.

        if provided or use that, otherwise update the current
        geometry to a new geometry format.


        Parameters
        ----------
        geometry : GeoSeries, optional
            New geometry, by default None
        geom_format: {'xy', 'wkt', 'geom'}
            Geometry format
        x_name, y_name, geom_name: str, optional
            The name of the x, y and geometry coordinate.
        replace : bool, optional
            Whether to replace the current geometry, by default True

        Returns
        -------
        Dataset, DataArray
            Update GeoDataset/Array
        """
        geom_format = geom_format or self.geom_format
        if geometry is None:
            geometry = self.geometry
        elif not isinstance(geometry, GeoSeries):
            raise ValueError("geometry should be a GeoSeries object")
        elif geometry.size != self.size:
            raise ValueError(
                f'The sizes of geometry and index dim "{self.index_dim}" do not match'
            )

        # update geometry and drop old
        drop_vars = []
        if self.geom_format != geom_format:
            if geom_format != "xy":
                drop_vars = [self.x_name, self.y_name]
            else:
                drop_vars = [self.geom_name]

        index_dim = self.index_dim
        if geom_format == "geom":
            if geom_name is None:
                geom_name = self.attrs.get("geom_name", "geometry")
            elif self.geom_name != geom_name:
                drop_vars.append(self.geom_name)
            coords = {geom_name: (self.index_dim, geometry.values)}
        elif geom_format == "wkt":
            if geom_name is None:
                geom_name = self.attrs.get("geom_name", "ogc_wkt")
            elif self.geom_name != geom_name:
                drop_vars.append(self.geom_name)
            coords = {geom_name: (self.index_dim, geometry.to_wkt().values)}
        elif geom_format == "xy":
            if x_name is None:
                x_name = self.attrs.get("x_name", "x")
            elif self.x_name != x_name:
                drop_vars.append(self.x_name)
            if y_name is None:
                y_name = self.attrs.get("y_name", "y")
            elif self.y_name != y_name:
                drop_vars.append(self.y_name)
            coords = {
                x_name: (self.index_dim, geometry.x.values),
                y_name: (self.index_dim, geometry.y.values),
            }
        obj = self._obj.copy()
        if replace:
            obj = obj.drop_vars(drop_vars, errors="ignore")
        obj = obj.assign_coords(coords)

        # reset spatial dims
        obj.vector._geometry = geometry
        obj.vector.set_spatial_dims(
            index_dim=index_dim,
            geom_format=geom_format,
            geom_name=geom_name,
            x_name=x_name,
            y_name=y_name,
        )
        obj.vector.set_crs(geometry.crs)
        return obj

    # Internal conversion and selection methods
    # i.e. produces xarray.Dataset/ xarray.DataArray
    def ogr_compliant(self, reducer=None) -> xr.Dataset:
        """Create a Dataset/Array which is understood by OGR.

        Variables with more than one dimension are not understood and will
        removed if no reducer is provided.

        Parameters
        ----------
        reducer : callable, optional
            method by which multidimensional data is reduced to 1 dimensional
            e.g. numpy.mean

        Returns
        -------
        xr.Dataset
            ogr compliant Dataset
        """
        obj = self.update_geometry(geom_format="wkt", geom_name="ogc_wkt")
        obj["ogc_wkt"].attrs = {
            "long_name": "Geometry as ISO WKT",
            "grid_mapping": "spatial_ref",
        }
        index_dim = self.index_dim

        if isinstance(self._obj, xr.DataArray):
            if obj.name in obj.coords:
                obj = obj.reset_coords(obj.name)
            else:
                if obj.name is None:
                    obj.name = "data"
                obj = obj.to_dataset()

        if reducer is not None:
            rdims = [dim for dim in obj.dims if dim != index_dim]
            obj = obj.reduce(reducer, dim=rdims)

        dtypes = {"i": "Integer64", "f": "Real", "U": "String"}
        for name, da in obj.data_vars.items():
            if index_dim not in da.dims:
                continue
            if reducer is not None:
                rdims = [dim for dim in obj.dims if dim != index_dim]
                obj[name] = obj[name].reduce(reducer, rdims)
            # set ogr meta data
            dtype = dtypes.get(obj[name].dtype.kind, None)
            if dtype is not None:
                obj[name].attrs.update(
                    {
                        "ogr_field_name": f"{name}",
                        "ogr_field_type": dtype,
                    }
                )
                if dtype == "String":
                    obj[name].attrs.update(
                        {
                            "ogr_field_width": 100,
                        }
                    )
        obj = obj.drop_vars("spatial_ref", errors="ignore")
        obj.vector.set_crs(self.crs)
        obj = obj.assign_attrs(
            {
                "Conventions": "CF-1.6",
                "GDAL": f"GDAL {GDAL_VERSION}",
                "ogr_geometry_field": "ogc_wkt",
                "ogr_layer_type": f"{self.geom_type}",
            }
        )
        return obj

    def to_geom(self, geom_name: str = None) -> Union[xr.DataArray, xr.Dataset]:
        """Convert Dataset/ DataArray with xy or wkt geometries to shapely Geometries.

        Parameters
        ----------
        geom_name: str, optional
            The name of the geometry coordinate.
            If None (default), derived from current object or default to 'geometry'.

        Returns
        -------
        xr.Dataset
            Dataset with new geometry coordinates
        """
        return self.update_geometry(geom_format="geom", geom_name=geom_name)

    def to_xy(self, x_name="x", y_name="y") -> Union[xr.DataArray, xr.Dataset]:
        """Convert Dataset/ DataArray with Point geometries to x,y structure.

        Parameters
        ----------
        x_name, y_name: str, optional
            The name of the x and y coordinates.

        Returns
        -------
        xr.Dataset
            Dataset with new x, y coordinates
        """
        return self.update_geometry(geom_format="xy", x_name=x_name, y_name=y_name)

    def to_wkt(
        self,
        ogr_compliant=False,
        reducer=None,
    ) -> Union[xr.DataArray, xr.Dataset]:
        """Convert geometries in Dataset/DataArray to wkt strings.

        Parameters
        ----------
        ogr_compliant: bool, optional
            Whether to create a OGR compliant Dataset/Array
            see :py:meth:`~hydromt.vector.GeoBase.ogr_compliant` for more details.
        reducer : callable, optional
            method by which multidimensional data is reduced to 1 dimensional
            e.g. numpy.mean

        Returns
        -------
        xr.Dataset
            Dataset with new ogc_wkt coordinate
        """
        # ogr compliant naming and attrs
        if ogr_compliant:
            obj = self.ogr_compliant(reducer=reducer)
        else:
            obj = self.update_geometry(geom_format="wkt", geom_name="ogc_wkt")
        return obj

    ## clip
    def clip_geom(self, geom, predicate="intersects"):
        """Select all geometries that intersect with the input geometry.

        Arguments
        ---------
        geom : geopandas.GeoDataFrame/Series,
            A geometry defining the area of interest.
        predicate : {None, 'intersects', 'within', 'contains', \
                     'overlaps', 'crosses', 'touches'}, optional
            If predicate is provided, the input geometry is tested
            using the predicate function against each item in the
            index whose extent intersects the envelope of the input geometry:
            predicate(input_geometry, tree_geometry).

        Returns
        -------
        da: xarray.DataArray
            Clipped DataArray
        """
        idx = gis_utils.filter_gdf(self.geometry, geom=geom, predicate=predicate)
        return self._obj.isel({self.index_dim: idx})

    def clip_bbox(self, bbox, crs=None, buffer=None):
        """Select point locations to bounding box.

        Arguments
        ---------
        bbox: tuple of floats
            (xmin, ymin, xmax, ymax) bounding box
        buffer: float, optional
            buffer around bbox in crs units, None by default.
        crs : int, optional
            EPSG of the data. If not given, it will be inferred.

        Returns
        -------
        da: xarray.DataArray
            Clipped DataArray
        """
        if buffer is not None:
            bbox = np.atleast_1d(bbox)
            bbox[:2] -= buffer
            bbox[2:] += buffer
        idx = gis_utils.filter_gdf(
            self.geometry, bbox=bbox, crs=crs, predicate="intersects"
        )
        return self._obj.isel({self.index_dim: idx})

    ## wrap GeoSeries functions
    # TODO write general wrapper
    def to_crs(self, dst_crs):
        """Transform spatial coordinates to a new coordinate reference system.

        The ``crs`` attribute on the current GeoDataArray must be set.

        Arguments
        ---------
        dst_crs: int, dict, or str, optional
            Accepts EPSG codes (int or str); proj (str or dict) or wkt (str)

        Returns
        -------
        xr.DataArray
            DataArray with transformed geospatial coordinates
        """
        if self.crs is None:
            raise ValueError("Source CRS is missing. Use da.vector.set_crs(crs) first.")
        geometry = self.geometry.to_crs(pyproj.CRS.from_user_input(dst_crs))
        return self.update_geometry(geometry)

    # Constructers
    # i.e. from other datatypes or files

    ## Output methods
    ## Either writes to files or other data types
    def to_gdf(self, reducer=None):
        """Return geopandas GeoDataFrame with Point geometry.

        Geometry is based on Dataset coordinates. If a reducer is
        passed the Dataset variables are reduced along the all
        non-index dimensions and to a GeoDataFrame column.

        Arguments
        ---------
        reducer : callable, str, optional
            method by which multidimensional data is reduced to 1 dimensional
            e.g. numpy.mean. If str (e.g. mean), will call the numpy method is found.

        Returns
        -------
        gdf: geopandas.GeoDataFrame
            GeoDataFrame
        """
        if isinstance(reducer, str):
            reducer = getattr(np, reducer)
        obj = self._obj
        if isinstance(obj, xr.DataArray):
            # we are looking at a coordinate
            if obj.name in obj.coords:
                obj = obj.reset_coords(obj.name)
            else:
                if obj.name is None:
                    obj.name = "data"
                obj = obj.to_dataset()
        gdf = obj.vector.geometry.to_frame("geometry")
        snames = ["y_name", "x_name", "index_dim", "geom_name"]
        sdims = [obj.vector.attrs.get(n) for n in snames if n in obj.vector.attrs]
        for name in obj.vector._all_names:
            dims = obj[name].dims
            if name not in sdims:
                # keep 1D variables with matching index_dim
                if len(dims) == 1 and dims[0] == obj.vector.index_dim:
                    gdf[name] = obj[name].values
                # keep reduced data variables
                elif reducer is not None and obj.vector.index_dim in obj[name].dims:
                    rdims = [
                        dim for dim in obj[name].dims if dim != obj.vector.index_dim
                    ]
                    gdf[name] = obj[name].reduce(reducer, dim=rdims)
        del obj
        return gdf

    def to_netcdf(
        self,
        path: str,
        ogr_compliant: bool = False,
        reducer=None,
        **kwargs,
    ) -> None:
        """Export geodataset vectordata to an ogr compliant netCDF4 file.

        Parameters
        ----------
        path : str
            Output path for netcdf file.
        ogr_compliant : bool
            write the netCDF4 file in an ogr compliant format
            This makes it readable as a vector file in e.g. QGIS
            see :py:meth:`~hydromt.vector.GeoBase.ogr_compliant` for more details.
        reducer : callable, optional
            Method by which multidimensional data is reduced to 1 dimensional
            e.g. numpy.mean
        kwargs:
            Any additional arguments to be passed down to the driver.
        """
        if ogr_compliant:
            self.ogr_compliant(reducer=reducer).to_netcdf(
                path, engine="netcdf4", **kwargs
            )
        else:
            obj = self.update_geometry(geom_format="wkt", geom_name="ogc_wkt")
            obj.to_netcdf(path, engine="netcdf4", **kwargs)
            del obj


@xr.register_dataarray_accessor("vector")
class GeoDataArray(GeoBase):

    """Accessor class for vector based geo data arrays."""

    def __init__(self, xarray_obj):
        """Initialise the object."""
        super(GeoDataArray, self).__init__(xarray_obj)

    @property
    def nodata(self):
        """Nodata value of the DataArray."""
        return self._obj.attrs.get("_FillValue", None)

    def set_nodata(self, nodata=None):
        """Set the nodata value of the DataArray.

        Parameters
        ----------
        nodata : float, int, opional
            nodata value
        """
        # Only numerical nodata values are supported
        if np.issubdtype(type(nodata), np.number):
            if not np.issubdtype(type(nodata), self._obj.dtype):
                nodata = self._obj.dtype.type(nodata)
            self._obj.attrs["_FillValue"] = nodata
        else:
            logger.warning("No numerical nodata value found, skipping set_nodata")
            self._obj.attrs.pop("_FillValue", None)

    # Constructers
    # i.e. from other datatypes or files
    @staticmethod
    def from_gdf(
        gdf: gpd.GeoDataFrame,
        data: Any,
        coords: dict = None,
        dims: tuple = None,
        name: str = None,
        index_dim: str = None,
        keep_cols: bool = True,
        merge_index: str = "gdf",
    ) -> xr.DataArray:
        """Parse GeoDataFrame object with point geometries to DataArray.

        DataArray will have geospatial attributes and be merged with array_like data.

        Arguments
        ---------
        gdf: geopandas GeoDataFrame
            Spatial coordinates. The index should match the array_like index_dim and the
            geometry column may only contain Point geometries.
        data: array_like
            Values for this array. Must be an ``numpy.ndarray``, ndarray like,
            or castable to an ``ndarray``. If a self-described xarray or pandas
            object, attempts are made to use this array's metadata to fill in
            other unspecified arguments. A view of the array's data is used
            instead of a copy if possible.
        coords: sequence or dict of array_like, optional
            Coordinates (tick labels) to use for indexing along each dimension.
        dims: hashable or sequence of hashable, optional
            Name(s) of the data dimension(s). Must be either a hashable (only
            for 1D data) or a sequence of hashables with length equal to the
            number of dimensions. If this argument is omitted, dimension names
            default to ``['dim_0', ... 'dim_n']``.
        name: str, optional
            The name of the data set for metadata purposes.
        index_dim: str, optional
            Name of index dimension of data
        keep_cols: bool, optional
            If True, keep gdf columns as extra coordinates in dataset
        merge_index: {'gdf', 'inner'}, default 'gdf'
            Type of merge to be performed between gdf and data.

            * gdf: use only keys from gdf index. Missing values will be filled with NaNs
            * inner: use intersection of gdf and data index.

        Returns
        -------
        da: xarray.DataArray
            DataArray with geospatial coordinates
        """
        # parse gdf to dataset
        if isinstance(gdf, GeoSeries):
            if gdf.name is None:
                gdf.name = "geometry"
            gdf = gdf.to_frame()
        if not isinstance(gdf, GeoDataFrame):
            raise ValueError(f"gdf data type not understood {type(gdf)}")
        # check index_dim
        if index_dim is None:
            index_dim = gdf.index.name if gdf.index.name is not None else "index"
        geom_name = gdf.geometry.name
        # create DataArray from array_like data
        da = xr.DataArray(data=data, coords=coords, dims=dims, name=name)
        # check dims -> assume index dim is first dim if not provided
        if dims is None and index_dim not in da.dims and len(da.dims) > 0:
            da = da.rename({list(da.dims)[0]: index_dim})
        # check if all data array contain index_dim
        if index_dim not in da.dims:
            raise ValueError(f"Index dimension {index_dim} not found on DataArray.")
        if np.dtype(da[index_dim]).type != np.dtype(gdf.index).type:
            try:
                da[index_dim] = da[index_dim].astype(np.dtype(gdf.index).type)
            except TypeError:
                raise TypeError(
                    "DataArray and GeoDataFrame index datatypes do not match."
                )
        # Which indices to use
        gdf_index = gdf.index.values
        ds_index = da[index_dim].values
        intersect_index = np.intersect1d(gdf_index, ds_index)
        if merge_index == "gdf":
            _index = gdf_index
        elif merge_index == "inner":
            _index = intersect_index
        else:
            raise ValueError(f"{merge_index} is not a valid value for 'merge_index'")
        if len(intersect_index) == 0:
            raise ValueError("No common indices found between gdf and data.")
        da = da.reindex({index_dim: _index}).transpose(index_dim, ...)
        # set gdf geometry and optional other columns
        hdrs = gdf.columns if keep_cols else [geom_name]
        da = da.assign_coords({hdr: (index_dim, gdf.loc[_index, hdr]) for hdr in hdrs})
        # set geospatial attributes
        da.vector.set_spatial_dims(geom_name=geom_name, geom_format="geom")
        da.vector.set_crs(gdf.crs)
        return da

    @staticmethod
    def from_netcdf(
        path: Union[str, xr.DataArray],
        parse_geom: bool = True,
        geom_name: str = None,
        x_name: str = None,
        y_name: str = None,
        crs: int = None,
        **kwargs,
    ) -> xr.DataArray:
        """Read netcdf file or convert xr.DataArray as GeoDataArray.

        Parameters
        ----------
        path : str, xr.DataArray
            path to file
        parse_geom : bool, optional
            Create geometry objects in place of existing x, y or wkt geometry
            coordinates.
        x_name, y_name, geom_name: str, optional
            The name of the x, y and geometry coordinate.
        crs : int, optional
            EPSG of the data. If not given, it will be inferred.
        **kwargs
            passed to :py:func:`xarray.open_dataarray`

        Returns
        -------
        xr.DataArray
            DataArray with vector as accessor
        """
        if isinstance(path, xr.DataArray):
            da = path
        else:
            da = xr.open_dataarray(path, **kwargs)
        da.vector.set_spatial_dims(geom_name=geom_name, x_name=x_name, y_name=y_name)
        # force to geom_format "geom"
        if parse_geom:
            da = da.vector.update_geometry(geom_format="geom", replace=True)
        da.vector.set_crs(input_crs=crs)  # try to parse from netcdf if None
        return da


@xr.register_dataset_accessor("vector")
class GeoDataset(GeoBase):

    """Implementation for a vectorased geo dataset."""

    def __init__(self, xarray_obj):
        """Initialise the object."""
        super(GeoDataset, self).__init__(xarray_obj)

    # Properties
    # Will probably be deleted in the future but now needed for compatibility
    @property
    def vars(self):
        """list: Returns non-coordinate varibles."""
        return list(self._obj.data_vars.keys())

    # Internal conversion and selection methods
    # i.e. produces xarray.Dataset/ xarray.DataArray

    # Constructers
    # i.e. from other datatypes or files
    @staticmethod
    def from_gdf(
        gdf: gpd.GeoDataFrame,
        data_vars: dict = None,
        coords: dict = None,
        index_dim: str = None,
        keep_cols: bool = True,
        cols_as_data_vars: bool = False,
        merge_index: str = "gdf",
    ) -> xr.Dataset:
        """Create Dataset with geospatial coordinates.

        Arguments
        ---------
        gdf: geopandas GeoDataFrame
            Spatial coordinates. The index should match the df index and the geometry
            columun may only contain Point geometries. Additional columns are also
            parsed to the xarray DataArray coordinates.
        data_vars: dict-like, DataArray or Dataset
            A mapping from variable names to `xarray.DataArray` objects.
            See `xarray.Dataset` for all options.
            Aditionally it accepts `xarray.DataArray` with name property and
            `xarray.Dataset`.
        coords: sequence or dict of array_like, optional
            Coordinates (tick labels) to use for indexing along each dimension.
        index_dim: str, optional
            Name of index dimension in data_vars
        keep_cols: bool, optional
            If True, keep gdf columns as extra coordinates in dataset
        cols_as_data_vars: bool, optional
            If True, parse gdf columns as data variables rather than coordinates.
        merge_index: {'gdf', 'inner'}, default 'gdf'
            Type of merge to be performed between gdf and data.

            * gdf: use only keys from gdf index. Missing values will be filled with NaNs
            * inner: use intersection of gdf and data index.

        Returns
        -------
        da: xarray.Dataset
            Dataset with geospatial coordinates
        """
        # parse gdf to dataset
        if isinstance(gdf, GeoSeries):
            if gdf.name is None:
                gdf.name = "geometry"
            gdf = gdf.to_frame()
        if not isinstance(gdf, GeoDataFrame):
            raise ValueError(f"gdf data type not understood {type(gdf)}")
        # check index_dim
        if index_dim is None:
            index_dim = gdf.index.name if gdf.index.name is not None else "index"
        geom_name = gdf.geometry.name
        # parse data_vars to dataset
        if data_vars is not None:
            if isinstance(data_vars, xr.DataArray):
                ds = data_vars.to_dataset()
            elif isinstance(data_vars, xr.Dataset):
                ds = data_vars
            else:
                try:
                    ds = xr.Dataset(data_vars, coords=coords)
                except TypeError:
                    raise TypeError(
                        "data_vars should be a dict-like, DataArray or Dataset"
                    )
            # check if any data array contain index_dim
            if index_dim not in ds.dims:
                raise ValueError(f"Index dimension {index_dim} not found in data_vars.")
            if np.dtype(ds[index_dim]).type != np.dtype(gdf.index).type:
                try:
                    ds[index_dim] = ds[index_dim].astype(np.dtype(gdf.index).type)
                except TypeError:
                    raise TypeError(
                        "Dataset and GeoDataFrame index datatypes do not match."
                    )
        else:  # create empty dataset
            ds = xr.Dataset(coords={index_dim: gdf.index.values})
        # Which indices to use
        gdf_index = gdf.index.values
        ds_index = ds[index_dim].values
        intersect_index = np.intersect1d(gdf_index, ds_index)
        if merge_index == "gdf":
            _index = gdf_index
        elif merge_index == "inner":
            _index = intersect_index
        else:
            raise ValueError(f"{merge_index} is not a valid value for 'merge_index'")
        if len(intersect_index) == 0:
            raise ValueError("No common indices found between gdf and data_vars.")
        ds = ds.reindex({index_dim: _index}).transpose(index_dim, ...)
        # set gdf geometry and optional other columns
        hdrs = gdf.columns if keep_cols else [geom_name]
        if cols_as_data_vars:
            for hdr in hdrs:
                if hdr != geom_name:
                    ds[hdr] = (index_dim, gdf.loc[_index, hdr])
                else:
                    ds = ds.assign_coords({hdr: (index_dim, gdf.loc[_index, hdr])})
        else:
            ds = ds.assign_coords(
                {hdr: (index_dim, gdf.loc[_index, hdr]) for hdr in hdrs}
            )
        # set geospatial attributes
        ds.vector.set_spatial_dims(geom_name=geom_name, geom_format="geom")
        ds.vector.set_crs(gdf.crs)
        return ds

    @staticmethod
    def from_netcdf(
        path: Union[str, xr.Dataset],
        parse_geom=True,
        geom_name=None,
        x_name=None,
        y_name=None,
        crs=None,
        **kwargs,
    ) -> xr.Dataset:
        """Create GeoDataset from ogr compliant netCDF4 file or xr.Dataset.

        Parameters
        ----------
        path : str, xr.Dataset
            Path to the netCDF4 file
        parse_geom : bool, optional
            Create geometry objects in place of existing x, y or
            wkt geometry coordinates.
        x_name, y_name, geom_name: str, optional
            The name of the x, y and geometry coordinate.
        crs : int, optional
            EPSG of the data. If not given, it will be inferred.
        **kwargs
            passed to :py:func:`xarray.open_dataset`

        Returns
        -------
        xr.Dataset
            Dataset containing the geospatial data and attributes
        """
        if isinstance(path, xr.Dataset):
            ds = path
        else:
            ds = xr.open_dataset(path, **kwargs)
        ds.vector.set_spatial_dims(geom_name=geom_name, x_name=x_name, y_name=y_name)
        # force geom_format= 'geom'
        if parse_geom:
            ds = ds.vector.update_geometry(geom_format="geom", replace=True)
        ds.vector.set_crs(crs)  # try to parse from netcdf if None
        return ds



# -*- coding: utf-8 -*-
"""HydroMT workflows."""
from .. import _compat
from .basin_mask import *
from .forcing import *
from .grid import *
from .rivers import *

if _compat.HAS_XUGRID:
    from .mesh import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""GIS related convenience functions."""
from __future__ import annotations

import glob
import logging
import os
from os.path import dirname
from typing import Optional, Tuple, Union

import geopandas as gpd
import numpy as np
import xarray as xr
from pyflwdir import gis_utils as gis
from pyogrio import read_info
from pyproj import CRS
from pyproj.transformer import Transformer
from rasterio.transform import Affine
from shapely.geometry import box
from shapely.geometry.base import BaseGeometry

from . import _compat

__all__ = ["spread2d", "nearest", "nearest_merge"]

logger = logging.getLogger(__name__)

_R = 6371e3  # Radius of earth in m. Use 3956e3 for miles
GDAL_DRIVER_CODE_MAP = {
    "asc": "AAIGrid",
    "blx": "BLX",
    "bmp": "BMP",
    "bt": "BT",
    "dat": "ZMap",
    "dem": "USGSDEM",
    "gen": "ADRG",
    "gif": "GIF",
    "gpkg": "GPKG",
    "grd": "NWT_GRD",
    "gsb": "NTv2",
    "gtx": "GTX",
    "hdr": "MFF",
    "hf2": "HF2",
    "hgt": "SRTMHGT",
    "img": "HFA",
    "jpg": "JPEG",
    "kro": "KRO",
    "lcp": "LCP",
    "mbtiles": "MBTiles",
    "mpr/mpl": "ILWIS",
    "ntf": "NITF",
    "pix": "PCIDSK",
    "png": "PNG",
    "pnm": "PNM",
    "rda": "R",
    "rgb": "SGI",
    "rst": "RST",
    "rsw": "RMF",
    "sdat": "SAGA",
    "sqlite": "Rasterlite",
    "ter": "Terragen",
    "tif": "GTiff",
    "vrt": "VRT",
    "xpm": "XPM",
    "xyz": "XYZ",
}
GDAL_EXT_CODE_MAP = {v: k for k, v in GDAL_DRIVER_CODE_MAP.items()}

GPD_TYPES = Union[gpd.GeoDataFrame, gpd.GeoSeries]
GEOM_TYPES = Union[GPD_TYPES, BaseGeometry]

## GEOM functions


def nearest_merge(
    gdf1: gpd.GeoDataFrame,
    gdf2: gpd.GeoDataFrame,
    columns: Optional[list] = None,
    max_dist: Optional[float] = None,
    overwrite: bool = False,
    inplace: bool = False,
    logger=logger,
) -> gpd.GeoDataFrame:
    """Merge attributes of gdf2 with the nearest feature of gdf1.

    Output is optionally bounded by a maximumum distance `max_dist`.
    Unless `overwrite = True`, gdf2 values are only
    merged where gdf1 has missing values.

    Parameters
    ----------
    gdf1, gdf2: geopandas.GeoDataFrame
        Source `gdf1` and destination `gdf2` geometries.
    columns : list of str, optional
        Names of columns in `gdf2` to merge, by default None
    max_dist : float, optional
        Maximum distance threshold for merge, by default None, i.e.: no threshold.
    overwrite : bool, optional
        If False (default) gdf2 values are only merged where gdf1 has missing values,
        i.e. NaN values for existing columns or missing columns.
    inplace : bool,
        If True, apply the merge to gdf1, otherwise return a new object.
    logger:
        The logger to use.

    Returns
    -------
    gpd.GeoDataFrame
        Merged GeoDataFrames
    """
    idx_nn, dst = nearest(gdf1, gdf2)
    if not inplace:
        gdf1 = gdf1.copy()
    valid = dst < max_dist if max_dist is not None else np.ones_like(idx_nn, dtype=bool)
    columns = gdf2.columns if columns is None else columns
    gdf1["distance_right"] = dst
    gdf1["index_right"] = -1
    gdf1.loc[valid, "index_right"] = idx_nn[valid]
    skip = ["geometry"]
    for col in columns:
        if col in skip or col not in gdf2:
            if col not in gdf2:
                logger.warning(f"Column {col} not found in gdf2 and skipped.")
            continue
        new_vals = gdf2.loc[idx_nn[valid], col].values
        if col in gdf1 and not overwrite:
            old_vals = gdf1.loc[valid, col]
            replace = np.logical_or(old_vals.isnull(), old_vals.eq(""))
            new_vals = np.where(replace, new_vals, old_vals)
        gdf1.loc[valid, col] = new_vals
    return gdf1


def nearest(
    gdf1: gpd.GeoDataFrame, gdf2: gpd.GeoDataFrame
) -> Tuple[np.ndarray, np.ndarray]:
    """Return the index of and distance [m] to the nearest geometry.

    For Line geometries in `gdf1` the nearest geometry is based line center point
    and for polygons on its representative point. Mixed geometry types are not
    yet supported.

    Note: Since geopandas v0.10.0 it contains a sjoin_nearest method which is very
    similar and should.


    Parameters
    ----------
    gdf1, gdf2: geopandas.GeoDataFrame
        Source `gdf1` and destination `gdf2` geometries.

    Returns
    -------
    index: ndarray
        index of nearest `gdf2` geometry
    dst: ndarray of float
        distance to the nearest `gdf2` geometry
    """
    if np.all(gdf1.type == "Point"):
        pnts = gdf1.geometry.copy()
    elif np.all(np.isin(gdf1.type, ["LineString", "MultiLineString"])):
        pnts = gdf1.geometry.interpolate(0.5, normalized=True)  # mid point
    elif np.all(np.isin(gdf1.type, ["Polygon", "MultiPolygon"])):
        pnts = gdf1.geometry.representative_point()  # inside polygon
    else:
        raise NotImplementedError("Mixed geometry dataframes are not yet supported.")
    if gdf1.crs != gdf2.crs:
        pnts = pnts.to_crs(gdf2.crs)
    # find nearest
    # NOTE: requires shapely v2.0; changed in v0.6.1
    if not _compat.HAS_SHAPELY20:
        raise ImportError("Shapely >= 2.0.0 is required for execution")
    other = pnts.geometry.values
    idx = gdf2.sindex.nearest(other, return_all=False)[1]
    # get distance in meters
    gdf2_nearest = gdf2.iloc[idx]
    if gdf2_nearest.crs.is_geographic:
        pnts = pnts.to_crs(3857)  # web mercator
        gdf2_nearest = gdf2_nearest.to_crs(3857)
    dst = gdf2_nearest.distance(pnts, align=False).values
    return gdf2.index.values[idx], dst


def filter_gdf(gdf, geom=None, bbox=None, crs=None, predicate="intersects"):
    """Filter GeoDataFrame geometries based on geometry mask or bounding box."""
    gtypes = (gpd.GeoDataFrame, gpd.GeoSeries, BaseGeometry)
    if bbox is not None and geom is None:
        if crs is None:
            crs = gdf.crs
        geom = gpd.GeoSeries([box(*bbox)], crs=crs)
    elif geom is not None and not isinstance(geom, gtypes):
        raise ValueError(
            f"Unknown geometry mask type {type(geom).__name__}. "
            "Provide geopandas GeoDataFrame, GeoSeries or shapely geometry."
        )
    elif bbox is None and geom is None:
        raise ValueError("Either geom or bbox is required.")
    if not isinstance(geom, BaseGeometry):
        # reproject
        if geom.crs is None and gdf.crs is not None:
            geom = geom.set_crs(gdf.crs)
        elif gdf.crs is not None and geom.crs != gdf.crs:
            geom = geom.to_crs(gdf.crs)
        # convert geopandas to geometry
        geom = geom.unary_union
    idx = np.sort(gdf.sindex.query(geom, predicate=predicate))
    return idx


def parse_geom_bbox_buffer(geom=None, bbox=None, buffer=0):
    """Parse geom or bbox to a (buffered) geometry.

    Arguments
    ---------
    geom : geopandas.GeoDataFrame/Series, optional
        A geometry defining the area of interest.
    bbox : array-like of floats, optional
        (xmin, ymin, xmax, ymax) bounding box of area of interest
        (in WGS84 coordinates).
    buffer : float, optional
        Buffer around the `bbox` or `geom` area of interest in meters. By default 0.

    Returns
    -------
    geom: geometry
        the actual geometry
    """
    if geom is None and bbox is not None:
        # convert bbox to geom with crs EPGS:4326 to apply buffer later
        geom = gpd.GeoDataFrame(geometry=[box(*bbox)], crs=4326)
    elif geom is None:
        raise ValueError("No geom or bbox provided.")

    if buffer > 0:
        # make sure geom is projected > buffer in meters!
        if geom.crs.is_geographic:
            geom = geom.to_crs(3857)
        geom = geom.buffer(buffer)
    return geom


# REPROJ
def utm_crs(bbox):
    """Return wkt string of nearest UTM projects.

    Parameters
    ----------
    bbox : array-like of floats
        (xmin, ymin, xmax, ymax) bounding box in latlon WGS84 (EPSG:4326) coordinates

    Returns
    -------
    crs: pyproj.CRS
        CRS of UTM projection
    """
    left, bottom, right, top = bbox
    x = (left + right) / 2
    y = (top + bottom) / 2
    kwargs = dict(zone=int(np.ceil((x + 180) / 6)))
    # BUGFIX hydroMT v0.3.5: south=False doesn't work only add south=True if y<0
    if y < 0:
        kwargs.update(south=True)
    # BUGFIX hydroMT v0.4.6: add datum
    epsg = CRS(proj="utm", datum="WGS84", ellps="WGS84", **kwargs).to_epsg()
    return CRS.from_epsg(epsg)


def parse_crs(crs, bbox=None):
    if crs == "utm":
        if bbox is not None:
            crs = utm_crs(bbox)
        else:
            raise ValueError('CRS "utm" requires bbox')
    else:
        crs = CRS.from_user_input(crs)
    return crs


def axes_attrs(crs):
    """Provide CF-compliant variable names and metadata for axes.

    Parameters
    ----------
    crs: pyproj.CRS
        coordinate reference system

    Returns
    -------
    x_dim: str - variable name of x dimension (e.g. 'x')
    y_dim: str - variable name of y dimension (e.g. 'lat')
    x_attr: dict - attributes of variable x
    y_attr: dict - attributes of variable y
    """
    # # check for type of crs
    if not isinstance(crs, CRS):
        crs = CRS.from_user_input(crs)
    if crs.is_geographic:
        x_dim, y_dim = "longitude", "latitude"
    else:
        x_dim, y_dim = "x", "y"
    cf_coords = crs.cs_to_cf()
    x_attrs = [c for c in cf_coords if c["axis"] == "X"][0]
    y_attrs = [c for c in cf_coords if c["axis"] == "Y"][0]
    return x_dim, y_dim, x_attrs, y_attrs


def meridian_offset(ds, bbox=None):
    """Shift data along the x-axis of global datasets to avoid issues along the 180 meridian.

    Without a bbox the data is shifted to span 180W to 180E.
    With bbox the data is shifted to at least span the bbox west to east,
    also if the bbox crosses the 180 meridian.

    Note that this method is only applicable to data that spans 360 degrees longitude
    and is set in a global geographic CRS (WGS84).

    Parameters
    ----------
    ds: xarray.Dataset
        input dataset
    bbox: tuple of float
        bounding box (west, south, east, north) in degrees

    Returns
    -------
    ds: xarray.Dataset
        dataset with x dim re-arranged if needed
    """
    w, _, e, _ = ds.raster.bounds
    if (
        ds.raster.crs is None
        or ds.raster.crs.is_projected
        or not np.isclose(e - w, 360)  # grid should span 360 degrees!
    ):
        raise ValueError(
            "This method is only applicable to data that spans 360 degrees "
            "longitude and is set in a global geographic CRS"
        )
    x_name = ds.raster.x_dim
    lons = np.copy(ds[x_name].values)
    if bbox is not None:  # bbox west and east
        bbox_w, bbox_e = bbox[0], bbox[2]
    else:  # global west and east in case of no bbox
        bbox_w, bbox_e = -180, 180
    if bbox_w < w:  # shift lons east of x0 by 360 degrees west
        x0 = 180 if bbox_w >= -180 else 0
        lons = np.where(lons > max(bbox_e, x0), lons - 360, lons)
    elif bbox_e > e:  # shift lons west of x0 by 360 degrees east
        x0 = -180 if bbox_e <= 180 else 0
        lons = np.where(lons < min(bbox_w, x0), lons + 360, lons)
    else:
        return ds
    ds = ds.copy(deep=False)  # make sure not to overwrite original ds
    ds[x_name] = xr.Variable(ds[x_name].dims, lons)
    return ds.sortby(x_name)


# TRANSFORM


def affine_to_coords(transform, shape, x_dim="x", y_dim="y"):
    """Return a raster axis with pixel center coordinates based on the transform.

    Parameters
    ----------
    transform : affine transform
        Two dimensional affine transform for 2D linear mapping
    shape : tuple of int
        The height, width  of the raster.
    x_dim, y_dim: str
        The name of the x and y dimensions

    Returns
    -------
    x, y coordinate arrays : dict of tuple with dims and coords
    """
    if not isinstance(transform, Affine):
        transform = Affine(*transform)
    height, width = shape
    if np.isclose(transform.b, 0) and np.isclose(transform.d, 0):
        x_coords, _ = transform * (np.arange(width) + 0.5, np.zeros(width) + 0.5)
        _, y_coords = transform * (np.zeros(height) + 0.5, np.arange(height) + 0.5)
        coords = {
            y_dim: (y_dim, y_coords),
            x_dim: (x_dim, x_coords),
        }
    else:
        x_coords, y_coords = (
            transform
            * transform.translation(0.5, 0.5)
            * np.meshgrid(np.arange(width), np.arange(height))
        )
        coords = {
            "yc": ((y_dim, x_dim), y_coords),
            "xc": ((y_dim, x_dim), x_coords),
        }
    return coords


def affine_to_meshgrid(transform, shape):
    """Return a meshgrid of pixel center coordinates based on the transform.

    Parameters
    ----------
    transform : affine transform
        Two dimensional affine transform for 2D linear mapping
    shape : tuple of int
        The height, width  of the raster.

    Returns
    -------
    x_coords, y_coords: ndarray
        2D arrays of x and y coordinates
    """
    if not isinstance(transform, Affine):
        transform = Affine(*transform)
    height, width = shape
    x_coords, y_coords = (
        transform
        * transform.translation(0.5, 0.5)
        * np.meshgrid(np.arange(width), np.arange(height))
    )
    return x_coords, y_coords


## CELLAREAS
def reggrid_area(lats, lons):
    """Return the cell area [m2] for a regular grid based on its cell centres lat, lon."""  # noqa: E501
    xres = np.abs(np.mean(np.diff(lons)))
    yres = np.abs(np.mean(np.diff(lats)))
    area = np.ones((lats.size, lons.size), dtype=lats.dtype)
    return cellarea(lats, xres, yres)[:, None] * area


def cellarea(lat, xres=1.0, yres=1.0):
    """Return the area [m2] of cell based on its center latitude and resolution in degrees.

    Resolution is in measured degrees.
    """  # noqa: E501
    l1 = np.radians(lat - np.abs(yres) / 2.0)
    l2 = np.radians(lat + np.abs(yres) / 2.0)
    dx = np.radians(np.abs(xres))
    return _R**2 * dx * (np.sin(l2) - np.sin(l1))


def cellres(lat, xres=1.0, yres=1.0):
    """Return the cell (x, y) resolution [m].

    Based on cell center latitude and its resolution measured in degrees.
    """
    m1 = 111132.92  # latitude calculation term 1
    m2 = -559.82  # latitude calculation term 2
    m3 = 1.175  # latitude calculation term 3
    m4 = -0.0023  # latitude calculation term 4
    p1 = 111412.84  # longitude calculation term 1
    p2 = -93.5  # longitude calculation term 2
    p3 = 0.118  # longitude calculation term 3

    radlat = np.radians(lat)  # numpy cos work in radians!
    # Calculate the length of a degree of latitude and longitude in meters
    dy = (
        m1
        + (m2 * np.cos(2.0 * radlat))
        + (m3 * np.cos(4.0 * radlat))
        + (m4 * np.cos(6.0 * radlat))
    )
    dx = (
        (p1 * np.cos(radlat))
        + (p2 * np.cos(3.0 * radlat))
        + (p3 * np.cos(5.0 * radlat))
    )

    return dx * xres, dy * yres


## SPREAD


def spread2d(
    da_obs: xr.DataArray,
    da_mask: Optional[xr.DataArray] = None,
    da_friction: Optional[xr.DataArray] = None,
    nodata: Optional[float] = None,
) -> xr.Dataset:
    """Return values of `da_obs` spreaded to cells with `nodata` value within `da_mask`.

    powered by :py:meth:`pyflwdir.gis_utils.spread2d`.

    Parameters
    ----------
    da_obs : xarray.DataArray
        Input raster with observation values and background/nodata values which are
        filled by the spreading algorithm.
    da_mask :  xarray.DataArray, optional
        Mask of cells to fill with the spreading algorithm, by default None
    da_friction :  xarray.DataArray, optional
        Friction values used by the spreading algorithm to calcuate the friction
        distance, by default None
    nodata : float, optional
        Nodata or background value. Must be finite numeric value. If not given the
        raster nodata value is used.

    Returns
    -------
    ds_out: xarray.Dataset
        Dataset with spreaded source values, linear index of the source cell
        "source_idx" and friction distance to the source cell "source_dst".
    """
    nodata = da_obs.raster.nodata if nodata is None else nodata
    if nodata is None or np.isnan(nodata):
        raise ValueError(f'"nodata" must be a finite value, not {nodata}')
    msk, frc = None, None
    if da_mask is not None:
        assert da_obs.raster.identical_grid(da_mask)
        msk = da_mask.values
    if da_friction is not None:
        assert da_obs.raster.identical_grid(da_friction)
        frc = da_friction.values
    out, src, dst = gis.spread2d(
        obs=da_obs.values,
        msk=msk,
        frc=frc,
        nodata=nodata,
        latlon=da_obs.raster.crs.is_geographic,
        transform=da_obs.raster.transform,
    )
    # combine outputs and return as dataset
    dims = da_obs.raster.dims
    coords = da_obs.raster.coords
    name = da_obs.name if da_obs.name else "source_value"
    da_out = xr.DataArray(dims=dims, coords=coords, data=out, name=name)
    da_out.raster.attrs.update(**da_obs.attrs)  # keep attrs incl nodata and unit
    da_src = xr.DataArray(dims=dims, coords=coords, data=src, name="source_idx")
    da_src.raster.set_nodata(-1)
    da_dst = xr.DataArray(dims=dims, coords=coords, data=dst, name="source_dst")
    da_dst.raster.set_nodata(-1)
    da_dst.attrs.update(unit="m")
    ds_out = xr.merge([da_out, da_src, da_dst])
    ds_out.raster.set_crs(da_obs.raster.crs)
    return ds_out


def create_vrt(
    vrt_path: str,
    files: list = None,
    files_path: str = None,
):
    r"""Create a .vrt file from a list op raster datasets.

    Either a list of files (`files`) or a path containing wildcards
    (`files_path`) to infer the list of files is required.

    Parameters
    ----------
    vrt_path : str
        Path of the output vrt
    files : list, optional
        List of raster datasets filenames, by default None
    files_path : str, optional
        Unix style path containing a pattern using wildcards (*)
        n.b. this is without an extension
        e.g. c:\\temp\\*\\*.tif for all tif files in subfolders of 'c:\temp'
    """
    if files is None and files_path is None:
        raise ValueError("Either 'files' or 'files_path' is required")

    if not _compat.HAS_RIO_VRT:
        raise ImportError(
            "rio-vrt is required for execution, install with 'pip install rio-vrt'"
        )
    import rio_vrt

    if files is None and files_path is not None:
        files = glob.glob(files_path)
        if len(files) == 0:
            raise IOError(f"No files found at {files_path}")

    outdir = dirname(vrt_path)
    if not os.path.isdir(outdir):
        os.makedirs(outdir)

    rio_vrt.build_vrt(vrt_path, files=files, relative=True)
    return None


def to_geographic_bbox(bbox, source_crs):
    target_crs = CRS.from_user_input(4326)
    if source_crs is None:
        logger.warning("No CRS was set. Skipping CRS conversion")
    elif source_crs != target_crs:
        bbox = Transformer.from_crs(source_crs, target_crs).transform_bounds(*bbox)

    return bbox


def bbox_from_file_and_filters(
    fn: str,
    bbox: Union[GEOM_TYPES, None] = None,
    mask: GEOM_TYPES | None = None,
    crs: CRS | None = None,
) -> Tuple[float, float, float, float] | None:
    """Create a bbox from the file metadata and filter options.

    Pyogrio does not accept a mask, and requires a bbox in the same CRS as the data.
    This function takes the possible bbox filter, mask filter and crs of the input data
    and returns a bbox in the same crs as the data based on the input filters.
    As pyogrio currently does not support filtering using a mask, the mask is converted
    to a bbox and the bbox is returned so that the data has some geospatial filtering.

    Parameters
    ----------
    fn: str,
        uri to the filename.
    bbox: GeoDataFrame | GeoSeries | BaseGeometry
        bounding box to filter the data while reading
    mask: GeoDataFrame | GeoSeries | BaseGeometry
        mask to filter the data while reading
    crs: pyproj.CRS
        coordinate reference system of the bounding box or geometry. If already set,
        this argument is ignored.
    """
    if bbox is not None and mask is not None:
        raise ValueError(
            "Both 'bbox' and 'mask' are provided. Please provide only one."
        )
    if bbox is None and mask is None:
        return None
    if source_crs_str := read_info(fn).get("crs"):
        source_crs = CRS(source_crs_str)
    elif crs:
        source_crs = crs
    else:  # assume WGS84
        source_crs = CRS("EPSG:4326")

    if mask is not None:
        bbox = mask

    # convert bbox to geom with input crs (assume WGS84 if not provided)
    crs = crs if crs is not None else CRS.from_user_input(4326)
    if issubclass(type(bbox), BaseGeometry):
        bbox = gpd.GeoSeries(bbox, crs=crs)
    bbox = bbox if bbox.crs is not None else bbox.set_crs(crs)
    return tuple(bbox.to_crs(source_crs).total_bounds)



"""Implementation of deprecation errors."""


class DeprecatedError(Exception):

    """Simple custom class to raise an error for something that is now deprecated."""

    def __init__(self, msg: str):
        """Initialise the object."""
        self.base = "DeprecationError"
        self.message = msg

    def __str__(self):
        return f"{self.base}: {self.message}"



# -*- coding: utf-8 -*-
"""Utils for parsing cli options and arguments."""

import json
import logging
from ast import literal_eval
from os.path import isfile
from pathlib import Path
from typing import Any, Dict, Optional, Union
from warnings import warn

import click
import yaml

from .. import config
from ..error import DeprecatedError

logger = logging.getLogger(__name__)

__all__ = ["parse_json", "parse_config", "parse_opt"]

### CLI callback methods ###


def parse_opt(ctx, param, value):
    """Parse extra cli options.

    Parse options like `--opt KEY1=VAL1 --opt SECT.KEY2=VAL2` and collect
    in a dictionary like the one below, which is what the CLI function receives.
    If no value or `None` is received then an empty dictionary is returned.
        {
            'KEY1': 'VAL1',
            'SECT': {
                'KEY2': 'VAL2'
                }
        }
    Note: `==VAL` breaks this as `str.split('=', 1)` is used.
    """
    out = {}
    if not value:
        return out
    for pair in value:
        if "=" not in pair:
            raise click.BadParameter("Invalid syntax for KEY=VAL arg: {}".format(pair))
        else:
            k, v = pair.split("=", 1)
            k = k.lower()
            s = None
            if "." in k:
                s, k = k.split(".", 1)
            try:
                v = literal_eval(v)
            except Exception:
                pass
            if s:
                if s not in out:
                    out[s] = dict()
                out[s].update({k: v})
            else:
                out.update({k: v})
    return out


def parse_json(ctx, param, value: str) -> Dict[str, Any]:
    """Parse json from object or file.

    If the object passed is a path pointing to a file, load it's contents and parse it.
    Otherwise attempt to parse the object as JSON itself.
    """
    if isfile(value):
        with open(value, "r") as f:
            kwargs = json.load(f)

    # Catch old keyword for resulution "-r"
    elif type(literal_eval(value)) in (float, int):
        raise DeprecatedError("'-r' is used for region, resolution is deprecated")
    else:
        if value.strip("{").startswith("'"):
            value = value.replace("'", '"')
        try:
            kwargs = json.loads(value)
        except json.JSONDecodeError:
            raise ValueError(f'Could not decode JSON "{value}"')
    return kwargs


### general parsing methods ##


def parse_config(
    path: Optional[Union[Path, str]] = None, opt_cli: Optional[Dict] = None
) -> Dict:
    """Parse config from ini `path` and combine with command line options `opt_cli`."""
    opt = {}
    if path is not None and isfile(path):
        if str(path).endswith(".ini"):
            warn(
                "Support for .ini configuration files will be deprecated",
                PendingDeprecationWarning,
                stacklevel=2,
            )
        opt = config.configread(
            path, abs_path=True, skip_abspath_sections=["setup_config"]
        )
    elif path is not None:
        raise IOError(f"Config not found at {path}")
    if opt_cli is not None:
        for section in opt_cli:
            if not isinstance(opt_cli[section], dict):
                raise ValueError(
                    "No section found in --opt values: "
                    "use <section>.<option>=<value> notation."
                )
            if section not in opt:
                opt[section] = opt_cli[section]
                continue
            for option, value in opt_cli[section].items():
                opt[section].update({option: value})
    return opt


def parse_export_config_yaml(ctx, param, value) -> Dict:
    if value:
        with open(value, "r") as stream:
            yml = yaml.load(stream, Loader=yaml.FullLoader)
        return yml
    else:
        return {}



# -*- coding: utf-8 -*-
"""General and basic API for models in HydroMT."""

import glob
import inspect
import logging
import os
import shutil
import typing
import warnings
from abc import ABCMeta
from os.path import abspath, basename, dirname, isabs, isdir, isfile, join
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr
from geopandas.testing import assert_geodataframe_equal
from pyproj import CRS
from shapely.geometry import box

from hydromt import __version__
from hydromt._compat import Distribution
from hydromt.typing import DeferedFileClose, StrPath, XArrayDict

from .. import config, log, workflows
from ..data_catalog import DataCatalog
from ..raster import GEO_MAP_COORD
from ..utils import _classproperty

__all__ = ["Model"]

logger = logging.getLogger(__name__)


class Model(object, metaclass=ABCMeta):

    """General and basic API for models in HydroMT."""

    _DATADIR = ""  # path to the model data folder
    _NAME = "modelname"
    _CONF = "model.ini"
    _CF = dict()  # configreader kwargs
    _GEOMS = {"<general_hydromt_name>": "<model_name>"}
    _MAPS = {"<general_hydromt_name>": "<model_name>"}
    _FOLDERS = [""]
    # tell hydroMT which methods should receive the res and region arguments
    # TODO: change it back to setup_region and no res --> deprecation
    _CLI_ARGS = {"region": "setup_basemaps", "res": "setup_basemaps"}
    _TMP_DATA_DIR = None
    # supported model version should be filled by the plugins
    # e.g. _MODEL_VERSION = ">=1.0, <1.1"
    _MODEL_VERSION = None

    _API = {
        "crs": CRS,
        "config": Dict[str, Any],
        "geoms": Dict[str, gpd.GeoDataFrame],
        "tables": Dict[str, pd.DataFrame],
        "maps": XArrayDict,
        "forcing": XArrayDict,
        "region": gpd.GeoDataFrame,
        "results": XArrayDict,
        "states": XArrayDict,
    }

    def __init__(
        self,
        root: Optional[str] = None,
        mode: Optional[str] = "w",
        config_fn: Optional[str] = None,
        data_libs: Union[List, str] = None,
        logger=logger,
        **artifact_keys,
    ):
        r"""Initialize a model.

        Parameters
        ----------
        root : str, optional
            Model root, by default None
        mode : {'r','r+','w'}, optional
            read/append/write mode, by default "w"
        config_fn : str, optional
            Model simulation configuration file, by default None.
            Note that this is not the HydroMT model setup configuration file!
        data_libs : List[str], optional
            List of data catalog configuration files, by default None
        \**artifact_keys:
            Additional keyword arguments to be passed down.
        logger:
            The logger to be used.
        """
        data_libs = data_libs or []
        from . import MODELS  # avoid circular import

        self.logger = logger
        dist_name, version = "unknown", "NA"
        if self._NAME in MODELS:
            ep = MODELS[self._NAME]
            dist: Distribution | None = ep.dist
            if dist:
                dist_name = dist.name
                version = dist.version
            else:
                # insert hydromt defaults
                dist_name = "hydromt"
                version = __version__

        # link to data
        self.data_catalog = DataCatalog(
            data_libs=data_libs, logger=self.logger, **artifact_keys
        )

        # placeholders
        # metadata maps that can be at different resolutions
        self._config: Optional[Dict[str:Literal]] = None  # nested dictionary
        self._maps: Optional[XArrayDict] = None
        self._tables: Dict[str, pd.DataFrame] = None

        self._geoms: Optional[Dict[str, gpd.GeoDataFrame]] = None
        self._forcing: Optional[XArrayDict] = None
        self._states: Optional[XArrayDict] = None
        self._results: Optional[XArrayDict] = None
        # To be deprecated in future versions!
        self._staticmaps = None
        self._staticgeoms = None

        # file system
        self._root = ""
        self._read = True
        self._write = False
        self._defered_file_closes = []

        # model paths
        self._config_fn = self._CONF if config_fn is None else config_fn
        self.set_root(root, mode)  # also creates hydromt.log file
        self.logger.info(
            f"Initializing {self._NAME} model from {dist_name} (v{version})."
        )

    @_classproperty
    def api(cls) -> Dict:
        """Return all model components and their data types."""
        _api = cls._API.copy()

        # reversed is so that child attributes take priority
        # this does mean that it becomes imporant in which order you
        # inherit from your base classes.
        for base_cls in reversed(cls.__mro__):
            if hasattr(base_cls, "_API"):
                _api.update(getattr(base_cls, "_API", {}))
        return _api

    def _check_get_opt(self, opt):
        """Check all opt keys and raise sensible error messages if unknown."""
        for method in opt.keys():
            m = method.strip("0123456789")
            if not callable(getattr(self, m, None)):
                raise ValueError(f'Model {self._NAME} has no method "{method}"')
        return opt

    def _run_log_method(self, method, *args, **kwargs):
        """Log method parameters before running a method."""
        method = method.strip("0123456789")
        func = getattr(self, method)
        signature = inspect.signature(func)
        # combine user and default options
        params = {}
        for i, (k, v) in enumerate(signature.parameters.items()):
            if k in ["args", "kwargs"]:
                if k == "args":
                    params[k] = args[i:]
                else:
                    params.update(**kwargs)
            else:
                v = kwargs.get(k, v.default)
                if len(args) > i:
                    v = args[i]
                params[k] = v
        # log options
        for k, v in params.items():
            if v is not inspect._empty:
                self.logger.info(f"{method}.{k}: {v}")
        return func(*args, **kwargs)

    def build(
        self,
        region: Optional[dict] = None,
        write: Optional[bool] = True,
        opt: Optional[dict] = None,
    ):
        r"""Single method to build a model from scratch based on settings in `opt`.

        Methods will be run one by one based on the order of appearance in `opt`
        (.ini configuration file). All model methods are supported including
        setup\_\*, read\_\* and write\_\* methods.

        If a write\_\* option is listed in `opt` (ini file) the full writing of the
        model at the end of the update process is skipped.

        Parameters
        ----------
        region: dict
            Description of model region. See :py:meth:`~hydromt.workflows.parse_region`
            for all options.
        write: bool, optional
            Write complete model after executing all methods in opt, by default True.
        opt: dict, optional
            Model build configuration. The configuration can be parsed from a
            .ini file using :py:meth:`~hydromt.config.configread`.
            This is a nested dictionary where the first-level keys are the names
            of model specific methods and the second-level contain
            argument-value pairs of the method.

            .. code-block:: text

                {
                    <name of method1>: {
                        <argument1>: <value1>, <argument2>: <value2>
                    },
                    <name of method2>: {
                        ...
                    }
                }

        """
        opt = opt or {}
        opt = self._check_get_opt(opt)

        # merge cli region and res arguments with opt
        if region is not None:
            if self._CLI_ARGS["region"] not in opt:
                opt = {self._CLI_ARGS["region"]: {}, **opt}
            opt[self._CLI_ARGS["region"]].update(region=region)

        # then loop over other methods
        for method in opt:
            # if any write_* functions are present in opt, skip the final self.write()
            if method.startswith("write_"):
                write = False
            kwargs = {} if opt[method] is None else opt[method]
            self._run_log_method(method, **kwargs)

        # write
        if write:
            self.write()

    def update(
        self,
        model_out: Optional[StrPath] = None,
        write: Optional[bool] = True,
        opt: Optional[Dict] = None,
        forceful_overwrite: bool = False,
    ):
        r"""Single method to update a model based the settings in `opt`.

        Methods will be run one by one based on the order of appearance in `opt`
        (ini configuration file).

        All model methods are supported including setup\_\*, read\_\* and write\_\* methods.
        If a write\_\* option is listed in `opt` (ini file) the full writing of the model
        at the end of the update process is skipped.

        Parameters
        ----------
        model_out: str, path, optional
            Destination folder to write the model schematization after updating
            the model. If None the updated model components are overwritten in the
            current model schematization if these exist. By default None.
        write: bool, optional
            Write the updated model schematization to disk. By default True.
        opt: dict, optional
            Model build configuration. The configuration can be parsed from a
            .ini file using :py:meth:`~hydromt.config.configread`.
            This is a nested dictionary where the first-level keys
            are the names of model specific methods and
            the second-level contain argument-value pairs of the method.

            .. code-block:: text

                {
                    <name of method1>: {
                        <argument1>: <value1>, <argument2>: <value2>
                    },
                    <name of method2>: {
                        ...
                    }
                }
          forceful_overwrite:
            Force open files to close when attempting to write them. In the case you
            try to write to a file that's already opened. The output will be written
            to a temporary file in case the original file cannot be written to.
        """
        opt = opt or {}
        opt = self._check_get_opt(opt)

        # read current model
        if not self._write:
            if model_out is None:
                raise ValueError(
                    '"model_out" directory required when updating in "read-only" mode'
                )
            self.read()
            if forceful_overwrite:
                self.set_root(model_out, mode="w+")
            else:
                self.set_root(model_out, mode="w")

        # check if model has a region
        if self.region is None:
            raise ValueError("Model region not found, setup model using `build` first.")

        # remove setup_basemaps from options and throw warning
        method = self._CLI_ARGS["region"]
        if method in opt:
            opt.pop(method)  # remove from opt
            self.logger.warning(f'"{method}" can only be called when building a model.')

        # loop over other methods from ini file
        for method in opt:
            # if any write_* functions are present in opt, skip the final self.write()
            if method.startswith("write_"):
                write = False
            kwargs = {} if opt[method] is None else opt[method]
            self._run_log_method(method, **kwargs)

        # write
        if write:
            self.write()

        self._cleanup(forceful_overwrite=forceful_overwrite)

    ## general setup methods

    def setup_region(
        self,
        region: dict,
        hydrography_fn: str = "merit_hydro",
        basin_index_fn: str = "merit_hydro_index",
    ) -> dict:
        """Set the `region` of interest of the model.

        Adds model layer:

        * **region** geom: region boundary vector

        Parameters
        ----------
        region : dict
            Dictionary describing region of interest, e.g.:

            * {'bbox': [xmin, ymin, xmax, ymax]}

            * {'geom': 'path/to/polygon_geometry'}

            * {'basin': [xmin, ymin, xmax, ymax]}

            * {'subbasin': [x, y], '<variable>': threshold}

            For a complete overview of all region options,
            see :py:function:~hydromt.workflows.basin_mask.parse_region
        hydrography_fn : str
            Name of data source for hydrography data.
            FIXME describe data requirements
        basin_index_fn : str
            Name of data source with basin (bounding box) geometries associated with
            the 'basins' layer of `hydrography_fn`. Only required if the `region` is
            based on a (sub)(inter)basins without a 'bounds' argument.

        Returns
        -------
        region: dict
            Parsed region dictionary

        See Also
        --------
        hydromt.workflows.basin_mask.parse_region
        """
        kind, region = workflows.parse_region(
            region, data_catalog=self.data_catalog, logger=self.logger
        )
        # NOTE: kind=outlet is deprecated!
        if kind in ["basin", "subbasin", "interbasin", "outlet"]:
            if kind == "outlet":
                warnings.warn(
                    "Using outlet as kind in setup_region is deprecated",
                    DeprecationWarning,
                    stacklevel=2,
                )
            # retrieve global hydrography data (lazy!)
            ds_org = self.data_catalog.get_rasterdataset(hydrography_fn)
            if "bounds" not in region:
                region.update(basin_index=self.data_catalog.get_source(basin_index_fn))
            # get basin geometry
            geom, xy = workflows.get_basin_geometry(
                ds=ds_org,
                kind=kind,
                logger=self.logger,
                **region,
            )
            region.update(xy=xy)
        elif "bbox" in region:
            bbox = region["bbox"]
            geom = gpd.GeoDataFrame(geometry=[box(*bbox)], crs=4326)
        elif "geom" in region:
            geom = region["geom"]
            if geom.crs is None:
                raise ValueError('Model region "geom" has no CRS')
        elif "grid" in region:  # Grid specific - should be removed in the future
            geom = region["grid"].raster.box
        elif "model" in region:
            geom = region["model"].region
        else:
            raise ValueError(f"model region argument not understood: {region}")

        self.set_geoms(geom, name="region")

        # This setup method returns region so that it can be wrapped for models which
        # require more information, e.g. grid RasterDataArray or xy coordinates.
        return region

    # TODO remove placeholder to make make sure
    # build with the current _CLI_ARGS does not raise an error
    def setup_basemaps(self, *args, **kwargs):  # noqa: D102
        warnings.warn(
            "The setup_basemaps method is not implemented.", UserWarning, stacklevel=2
        )

    ## file system

    @property
    def root(self):
        """Path to model folder."""
        if self._root is None:
            raise ValueError("Root unknown, use set_root method")
        return self._root

    def _assert_write_mode(self):
        if not self._write:
            raise IOError("Model opened in read-only mode")

    def _assert_read_mode(self):
        if not self._read:
            raise IOError("Model opened in write-only mode")

    def set_root(self, root: Optional[str], mode: Optional[str] = "w"):
        """Initialize the model root.

        In read/append mode a check is done if the root exists.
        In write mode the required model folder structure is created.

        Parameters
        ----------
        root : str, optional
            path to model root
        mode : {"r", "r+", "w"}, optional
            read/append/write mode for model files
        """
        ignore_ext = set([".log", ".yml"])
        if mode not in ["r", "r+", "w", "w+"]:
            raise ValueError(
                f'mode "{mode}" unknown, select from "r", "r+", "w" or "w+"'
            )
        self._root = root if root is None else abspath(root)
        self._read = mode.startswith("r")
        self._write = mode != "r"
        self._overwrite = mode == "w+"
        if self._root is not None:
            if self._write:
                for name in self._FOLDERS:
                    path = join(self._root, name)
                    if not isdir(path):
                        os.makedirs(path)
                        continue
                    # path already exists check files
                    fns = glob.glob(join(path, "*.*"))
                    exts = set([os.path.splitext(fn)[1] for fn in fns])
                    exts -= ignore_ext
                    if len(exts) != 0:
                        if mode.endswith("+"):
                            self.logger.warning(
                                "Model dir already exists and "
                                f"files might be overwritten: {path}."
                            )
                        else:
                            msg = (
                                "Model dir already exists and cannot be "
                                + f"overwritten: {path}. Use 'mode=w+' to force "
                                + "overwrite existing files."
                            )
                            self.logger.error(msg)
                            raise IOError(msg)
            # check directory
            elif not isdir(self._root):
                raise IOError(f'model root not found at "{self._root}"')
            # remove old logging file handler and add new filehandler
            # in root if it does not exist
            has_log_file = False
            log_level = 20  # default, but overwritten by the level of active loggers
            for i, h in enumerate(self.logger.handlers):
                log_level = h.level
                if hasattr(h, "baseFilename"):
                    if dirname(h.baseFilename) != self._root:
                        # remove handler and close file
                        self.logger.handlers.pop(i).close()
                    else:
                        has_log_file = True
                    break
            if not has_log_file:
                new_path = join(self._root, "hydromt.log")
                log.add_filehandler(self.logger, new_path, log_level)

    # I/O
    def read(
        self,
        components: List = None,
    ) -> None:
        """Read the complete model schematization and configuration from model files.

        Parameters
        ----------
        components : List, optional
            List of model components to read, each should have an associated
            read_<component> method. By default ['config', 'maps', 'staticmaps',
            'geoms', 'forcing', 'states', 'results']
        """
        if components is None:
            components = [
                "config",
                "staticmaps",
                "maps",
                "tables",
                "geoms",
                "forcing",
                "states",
                "results",
            ]
        self.logger.info(f"Reading model data from {self.root}")
        for component in components:
            if not hasattr(self, f"read_{component}"):
                raise AttributeError(
                    f"{type(self).__name__} does not have read_{component}"
                )
            getattr(self, f"read_{component}")()

    def write(
        self,
        components: List = None,
    ) -> None:
        """Write the complete model schematization and configuration to model files.

        Parameters
        ----------
        components : List, optional
            List of model components to write, each should have an
            associated write_<component> method. By default ['config', 'maps',
            'staticmaps', 'geoms', 'forcing', 'states']
        """
        if components is None:
            components = [
                "staticmaps",
                "maps",
                "tables",
                "geoms",
                "forcing",
                "states",
                "config",
            ]
        self.logger.info(f"Writing model data to {self.root}")
        for component in components:
            if not hasattr(self, f"write_{component}"):
                raise AttributeError(
                    f"{type(self).__name__} does not have write_{component}"
                )
            getattr(self, f"write_{component}")()

    def write_data_catalog(
        self,
        root: Optional[StrPath] = None,
        data_lib_fn: StrPath = "hydromt_data.yml",
        used_only: bool = True,
        append: bool = True,
        save_csv: bool = False,
    ):
        """Write the data catalog to data_lib_fn.

        Parameters
        ----------
        root: str, Path, optional
            Global root for all relative paths in configuration file.
            If "auto" the data source paths are relative to the yaml output ``path``.
        data_lib_fn: str, Path, optional
            Path of output yml file, absolute or relative to the model root,
            by default "hydromt_data.yml".
        used_only: bool, optional
            If True, export only data entries kept in used_data list. By default True
        append: bool, optional
            If True, append to an existing
        save_csv: bool, optional
            If True, save the data catalog also as an csv table. By default False.
        """
        path = data_lib_fn if isabs(data_lib_fn) else join(self.root, data_lib_fn)
        cat = DataCatalog(logger=self.logger, fallback_lib=None)
        # read hydromt_data yml file and add to data catalog
        if self._read and isfile(path) and append:
            cat.from_yml(path)
        # update data catalog with new used sources
        for name, source in self.data_catalog.iter_sources(used_only=used_only):
            cat.add_source(name, source)
        # write data catalog
        if cat.sources:
            self._assert_write_mode()
            if save_csv:
                csv_path = os.path.splitext(path)[0] + ".csv"
                cat.to_dataframe().reset_index().to_csv(
                    csv_path, sep=",", index=False, header=True
                )

            cat.to_yml(path, root=root)

    # model configuration
    @property
    def config(self) -> Dict[str, Union[Dict, str]]:
        """Model configuration. Returns a (nested) dictionary."""
        if self._config is None:
            self._initialize_config()
        return self._config

    def _initialize_config(self, skip_read=False) -> None:
        """Initialize the model config."""
        if self._config is None:
            self._config = dict()
            if not skip_read:
                # no check for read mode here
                # model config is read if in read-mode and it exists
                # default config if in write-mode
                self.read_config()

    def set_config(self, *args):
        """Update the config dictionary at key(s) with values.

        Parameters
        ----------
        args : key(s), value tuple, with minimal length of two
            keys can given by multiple args: ('key1', 'key2', 'value')
            or a string with '.' indicating a new level: ('key1.key2', 'value')

        Examples
        --------
        >> # self.config = {'a': 1, 'b': {'c': {'d': 2}}}

        >> set_config('a', 99)
        >> {'a': 99, 'b': {'c': {'d': 2}}}

        >> set_config('b', 'c', 'd', 99) # identical to set_config('b.d.e', 99)
        >> {'a': 1, 'b': {'c': {'d': 99}}}
        """
        self._initialize_config()
        if len(args) < 2:
            raise TypeError("set_config() requires a least one key and one value.")
        args = list(args)
        value = args.pop(-1)
        if len(args) == 1 and "." in args[0]:
            args = args[0].split(".") + args[1:]
        branch = self._config
        for key in args[:-1]:
            if key not in branch or not isinstance(branch[key], dict):
                branch[key] = {}
            branch = branch[key]
        branch[args[-1]] = value

    def read_config(self, config_fn: Optional[str] = None):
        """Parse config from file.

        If no config file found a default config file is returned in writing mode.
        """
        prefix = "User defined"
        if config_fn is None:  # prioritize user defined config path (new v0.4.1)
            if not self._read:  # write-only mode > read default config
                config_fn = join(self._DATADIR, self._NAME, self._CONF)
                prefix = "Default"
            elif self.root is not None:  # append or write mode > read model config
                config_fn = join(self.root, self._config_fn)
                prefix = "Model"
        cfdict = dict()
        if config_fn is not None:
            if isfile(config_fn):
                cfdict = self._configread(config_fn)
                self.logger.debug(f"{prefix} config read from {config_fn}")
            elif (
                self._root is not None
                and not isabs(config_fn)
                and isfile(join(self._root, config_fn))
            ):
                cfdict = self._configread(join(self.root, config_fn))
                self.logger.debug(
                    f"{prefix} config read from {join(self.root,config_fn)}"
                )
            elif isfile(abspath(config_fn)):
                cfdict = self._configread(abspath(config_fn))
                self.logger.debug(f"{prefix} config read from {abspath(config_fn)}")
            else:  # skip for missing default
                self.logger.error(f"{prefix} config file not found at {config_fn}")

        # always overwrite config when reading
        self._config = cfdict

    def write_config(
        self, config_name: Optional[str] = None, config_root: Optional[str] = None
    ):
        """Write config to <root/config_fn>."""
        self._assert_write_mode()
        if config_name is not None:
            self._config_fn = config_name
        elif self._config_fn is None:
            self._config_fn = self._CONF
        if config_root is None:
            config_root = self.root
        fn = join(config_root, self._config_fn)
        self.logger.info(f"Writing model config to {fn}")
        self._configwrite(fn)

    def _configread(self, fn: str):
        return config.configread(fn, abs_path=False)

    def _configwrite(self, fn: str):
        return config.configwrite(fn, self.config)

    def setup_config(self, **cfdict):
        """Update config with a dictionary."""
        # TODO rename to update_config
        if len(cfdict) > 0:
            self.logger.debug("Setting model config options.")
        for key, value in cfdict.items():
            self.set_config(key, value)

    def get_config(self, *args, fallback=None, abs_path: Optional[bool] = False):
        """Get a config value at key(s).

        Parameters
        ----------
        args : tuple or string
            keys can given by multiple args: ('key1', 'key2')
            or a string with '.' indicating a new level: ('key1.key2')
        fallback: any, optional
            fallback value if key(s) not found in config, by default None.
        abs_path: bool, optional
            If True return the absolute path relative to the model root,
            by deafult False.
            NOTE: this assumes the config is located in model root!

        Returns
        -------
        value : any type
            dictionary value

        Examples
        --------
        >> # self.config = {'a': 1, 'b': {'c': {'d': 2}}}

        >> get_config('a')
        >> 1

        >> get_config('b', 'c', 'd') # identical to get_config('b.c.d')
        >> 2

        >> get_config('b.c') # # identical to get_config('b','c')
        >> {'d': 2}
        """
        args = list(args)
        if len(args) == 1 and "." in args[0]:
            args = args[0].split(".") + args[1:]
        branch = self.config  # reads config at first call
        for key in args[:-1]:
            branch = branch.get(key, {})
            if not isinstance(branch, dict):
                branch = dict()
                break
        value = branch.get(args[-1], fallback)
        if abs_path and isinstance(value, str):
            value = Path(value)
            if not value.is_absolute():
                value = Path(abspath(join(self.root, value)))
        return value

    @property
    def tables(self) -> Dict[str, pd.DataFrame]:
        """Model tables."""
        if self._tables is None:
            self._initialize_tables()
        return self._tables

    def _initialize_tables(self, skip_read=False) -> None:
        """Initialize the model tables."""
        if self._tables is None:
            self._tables = dict()
            if self._read and not skip_read:
                self.read_tables()

    def write_tables(self, fn: str = "tables/{name}.csv", **kwargs) -> None:
        """Write tables at <root>/tables."""
        if self.tables:
            self._assert_write_mode()
            self.logger.info("Writing table files.")
            local_kwargs = {"index": False, "header": True, "sep": ","}
            local_kwargs.update(**kwargs)
            for name in self.tables:
                fn_out = join(self.root, fn.format(name=name))
                os.makedirs(dirname(fn_out), exist_ok=True)
                self.tables[name].to_csv(fn_out, **local_kwargs)
        else:
            self.logger.debug("No tables found, skip writing.")

    def read_tables(self, fn: str = "tables/{name}.csv", **kwargs) -> None:
        """Read table files at <root>/tables and parse to dict of dataframes."""
        self._assert_read_mode()
        self._initialize_tables(skip_read=True)
        self.logger.info("Reading model table files.")
        fns = glob.glob(join(self.root, fn.format(name="*")))
        if len(fns) > 0:
            for fn in fns:
                name = basename(fn).split(".")[0]
                tbl = pd.read_csv(fn, **kwargs)
                self.set_tables(tbl, name=name)

    def set_tables(
        self, tables: Union[pd.DataFrame, pd.Series, Dict], name=None
    ) -> None:
        """Add (a) table(s) <pandas.DataFrame> to model.

        Parameters
        ----------
        tables : pandas.DataFrame, pandas.Series or dict
            Table(s) to add to model.
            Multiple tables can be added at once by passing a dict of tables.
        name : str, optional
            Name of table, by default None. Required when tables is not a dict.
        """
        self._initialize_tables()
        if not isinstance(tables, dict) and name is None:
            raise ValueError("name required when tables is not a dict")
        elif not isinstance(tables, dict):
            tables = {name: tables}
        for name, df in tables.items():
            if not (isinstance(df, pd.DataFrame) or isinstance(df, pd.Series)):
                raise ValueError(
                    "table type not recognized, should be pandas DataFrame or Series."
                )
            if name in self._tables:
                if not self._write:
                    raise IOError(f"Cannot overwrite table {name} in read-only mode")
                elif self._read:
                    self.logger.warning(f"Overwriting table: {name}")

            self._tables[name] = df

    def get_tables_merged(self) -> pd.DataFrame:
        """Return all tables of a model merged into one dataframe."""
        # This is mostly used for convenience and testing.
        return pd.concat(
            [df.assign(table_origin=name) for name, df in self.tables.items()], axis=0
        )

    # model static maps
    @property
    def staticmaps(self):
        """Model static maps.

        Returns xarray.Dataset,
        ..NOTE: will be deprecated in future versions and replaced by `grid`.
        """
        warnings.warn(
            "The staticmaps property of the Model class will be deprecated in future"
            "versions. Use the grid property of the GridModel class instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        if self._staticmaps is None:
            self._staticmaps = xr.Dataset()
            if self._read:
                self.read_staticmaps()
        return self._staticmaps

    def set_staticmaps(
        self, data: Union[xr.DataArray, xr.Dataset], name: Optional[str] = None
    ):
        """Add data to staticmaps.

        All layers of staticmaps must have identical spatial coordinates.
        This method will be deprecated in future versions. See
        :py:meth:`~hydromt.models.GridModel.set_grid`.

        Parameters
        ----------
        data: xarray.DataArray or xarray.Dataset
            new map layer to add to staticmaps
        name: str, optional
            Name of new map layer, this is used to overwrite the name of a DataArray
            or to select a variable from a Dataset.
        """
        warnings.warn(
            "The set_staticmaps method will be deprecated in future versions, "
            + "use set_grid instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        if name is None:
            if isinstance(data, xr.DataArray) and data.name is not None:
                name = data.name
            elif not isinstance(data, xr.Dataset):
                raise ValueError("Setting a map requires a name")
        elif name is not None and isinstance(data, xr.Dataset):
            data_vars = list(data.data_vars)
            if len(data_vars) == 1 and name not in data_vars:
                data = data.rename_vars({data_vars[0]: name})
            elif name not in data_vars:
                raise ValueError("Name not found in DataSet")
            else:
                data = data[[name]]
        if isinstance(data, xr.DataArray):
            data.name = name
            data = data.to_dataset()
        if len(self.staticmaps) == 0:  # new data / trigger read
            self._staticmaps = data
        else:
            if isinstance(data, np.ndarray):
                if data.shape != self.shape:
                    raise ValueError("Shape of data and staticmaps do not match")
                data = xr.DataArray(dims=self.dims, data=data, name=name).to_dataset()
            for dvar in data.data_vars.keys():
                if dvar in self._staticmaps:
                    self.logger.warning(f"Replacing staticmap: {dvar}")
                self._staticmaps[dvar] = data[dvar]

    def read_staticmaps(self, fn: str = "staticmaps/staticmaps.nc", **kwargs) -> None:
        r"""Read static model maps at <root>/<fn> and add to staticmaps property.

        key-word arguments are passed to :py:meth:`~hydromt.models.Model.read_nc`

        .. NOTE: this method is deprecated.
        Use the grid property of the GridMixin instead.

        Parameters
        ----------
        fn : str, optional
            filename relative to model root, by default "staticmaps/staticmaps.nc"
        \**kwargs:
            Additional keyword arguments that are passed to the
            `read_nc` function.
        """
        self._assert_read_mode()
        for ds in self.read_nc(fn, **kwargs).values():
            self.set_staticmaps(ds)

    def write_staticmaps(self, fn: str = "staticmaps/staticmaps.nc", **kwargs) -> None:
        r"""Write static model maps to netcdf file at <root>/<fn>.

        key-word arguments are passed to :py:meth:`~hydromt.models.Model.write_nc`

        .. NOTE: this method is deprecated.
        Use the grid property of the GridMixin instead.

        Parameters
        ----------
        fn : str, optional
            filename relative to model root, by default 'staticmaps/staticmaps.nc'
        \**kwargs:
            Additional keyword arguments that are passed to the
            `write_nc` function.
        """
        if len(self.staticmaps) == 0:
            self.logger.debug("No staticmaps data found, skip writing.")
        else:
            self._assert_write_mode()
            # write_nc requires dict - use dummy 'staticmaps' key
            nc_dict = {"staticmaps": self.staticmaps}
            self.write_nc(nc_dict, fn, **kwargs)

    # map files setup methods
    def setup_maps_from_rasterdataset(
        self,
        raster_fn: Union[str, Path, xr.Dataset],
        variables: Optional[List] = None,
        fill_method: Optional[str] = None,
        name: Optional[str] = None,
        reproject_method: Optional[str] = None,
        split_dataset: Optional[bool] = True,
        rename: Optional[Dict] = None,
    ) -> List[str]:
        """HYDROMT CORE METHOD: Add data variable(s) from ``raster_fn`` to maps object.

        If raster is a dataset, all variables will be added unless ``variables``
        list is specified.

        Adds model layers:

        * **raster.name** maps: data from raster_fn

        Parameters
        ----------
        raster_fn: str, Path, xr.Dataset
            Data catalog key, path to raster file or raster xarray data object.
        variables: list, optional
            List of variables to add to maps from raster_fn. By default all.
        fill_method : str, optional
            If specified, fills nodata values using fill_nodata method.
            Available methods are {'linear', 'nearest', 'cubic', 'rio_idw'}.
        name: str, optional
            Name of new dataset in self.maps dictionnary,
            only in case split_dataset=False.
        reproject_method: str, optional
            See rasterio.warp.reproject for existing methods, by default the data is
            not reprojected (None).
        split_dataset: bool, optional
            If data is a xarray.Dataset split it into several xarray.DataArrays.
        rename: dict, optional
            Dictionary to rename variable names in raster_fn before adding to maps
            {'name_in_raster_fn': 'name_in_maps'}. By default empty.

        Returns
        -------
        list
            Names of added model map layers
        """
        rename = rename or {}
        self.logger.info(f"Preparing maps data from raster source {raster_fn}")
        # Read raster data and select variables
        ds = self.data_catalog.get_rasterdataset(
            raster_fn,
            geom=self.region,
            buffer=2,
            variables=variables,
            single_var_as_array=False,
        )
        # Fill nodata
        if fill_method is not None:
            ds = ds.raster.interpolate_na(method=fill_method)
        # Reprojection
        if ds.rio.crs != self.crs and reproject_method is not None:
            ds = ds.raster.reproject(dst_crs=self.crs, method=reproject_method)
        # Rename and add to maps
        self.set_maps(ds.rename(rename), name=name, split_dataset=split_dataset)

        return list(ds.data_vars.keys())

    def setup_maps_from_raster_reclass(
        self,
        raster_fn: Union[str, Path, xr.DataArray],
        reclass_table_fn: Union[str, Path, pd.DataFrame],
        reclass_variables: List,
        variable: Optional[str] = None,
        fill_method: Optional[str] = None,
        reproject_method: Optional[str] = None,
        name: Optional[str] = None,
        split_dataset: Optional[bool] = True,
        rename: Optional[Dict] = None,
        **kwargs,
    ) -> List[str]:
        r"""HYDROMT CORE METHOD: Add data variable(s) to maps object by reclassifying the data in ``raster_fn`` based on ``reclass_table_fn``.

        This is done by reclassifying the data in
        ``raster_fn`` based on ``reclass_table_fn``.

        Adds model layers:

        * **reclass_variables** maps: reclassified raster data

        Parameters
        ----------
        raster_fn: str, Path, xr.DataArray
            Data catalog key, path to raster file or raster xarray data object.
            Should be a DataArray. Else use `variable` argument for selection.
        reclass_table_fn: str, Path, pd.DataFrame
            Data catalog key, path to tabular data file or tabular pandas dataframe
            object for the reclassification table of `raster_fn`.
        reclass_variables: list
            List of reclass_variables from reclass_table_fn table to add to maps. Index
            column should match values in `raster_fn`.
        variable: str, optional
            Name of raster dataset variable to use. This is only required when reading
            datasets with multiple variables. By default None.
        fill_method : str, optional
            If specified, fills nodata values in `raster_fn` using fill_nodata method
            before reclassifying. Available methods are {'linear', 'nearest',
            'cubic', 'rio_idw'}.
        reproject_method: str, optional
            See rasterio.warp.reproject for existing methods, by default the data is
            not reprojected (None).
        name: str, optional
            Name of new maps variable, only in case split_dataset=False.
        split_dataset: bool, optional
            If data is a xarray.Dataset split it into several xarray.DataArrays.
        rename: dict, optional
            Dictionary to rename variable names in reclass_variables before adding to
            grid {'name_in_reclass_table': 'name_in_grid'}. By default empty.
        \**kwargs:
            Additional keyword arguments that are passed to the
            `data_catalog.get_rasterdataset` function.

        Returns
        -------
        list
            Names of added model map layers
        """  # noqa: E501
        rename = rename or {}
        self.logger.info(
            f"Preparing map data by reclassifying the data in {raster_fn} based"
            f" on {reclass_table_fn}"
        )
        # Read raster data and remapping table
        da = self.data_catalog.get_rasterdataset(
            raster_fn, geom=self.region, buffer=2, variables=variable, **kwargs
        )
        if not isinstance(da, xr.DataArray):
            raise ValueError(
                f"raster_fn {raster_fn} should be a single variable. "
                "Please select one using the 'variable' argument"
            )
        df_vars = self.data_catalog.get_dataframe(
            reclass_table_fn, variables=reclass_variables
        )
        # Fill nodata
        if fill_method is not None:
            da = da.raster.interpolate_na(method=fill_method)
        # Mapping function
        ds_vars = da.raster.reclassify(reclass_table=df_vars, method="exact")
        # Reprojection
        if ds_vars.rio.crs != self.crs and reproject_method is not None:
            ds_vars = ds_vars.raster.reproject(dst_crs=self.crs)
        # Add to maps
        self.set_maps(ds_vars.rename(rename), name=name, split_dataset=split_dataset)

        return list(ds_vars.data_vars.keys())

    # model map
    @property
    def maps(self) -> Dict[str, Union[xr.Dataset, xr.DataArray]]:
        """Model maps. Returns dict of xarray.DataArray or xarray.Dataset."""
        if self._maps is None:
            self._initialize_maps()
        return self._maps

    def _initialize_maps(self, skip_read=False) -> None:
        """Initialize maps."""
        if self._maps is None:
            self._maps = dict()
            if self._read and not skip_read:
                self.read_maps()

    def set_maps(
        self,
        data: Union[xr.DataArray, xr.Dataset],
        name: Optional[str] = None,
        split_dataset: Optional[bool] = True,
    ) -> None:
        """Add raster data to the maps component.

        Dataset can either be added as is (default) or split into several
        DataArrays using the split_dataset argument.

        Arguments
        ---------
        data: xarray.Dataset or xarray.DataArray
            New forcing data to add
        name: str, optional
            Variable name, only in case data is of type DataArray or if a Dataset is
            added as is (split_dataset=False).
        split_dataset: bool, optional
            If data is a xarray.Dataset split it into several xarray.DataArrays.
        """
        self._initialize_maps()
        data_dict = _check_data(data, name, split_dataset)
        for name in data_dict:
            if name in self._maps:
                self.logger.warning(f"Replacing result: {name}")
            self._maps[name] = data_dict[name]

    def read_maps(self, fn: str = "maps/*.nc", **kwargs) -> None:
        r"""Read model map at <root>/<fn> and add to maps component.

        key-word arguments are passed to :py:meth:`~hydromt.models.Model.read_nc`

        Parameters
        ----------
        fn : str, optional
            filename relative to model root, may contain wildcards,
            by default ``maps/\*.nc``
        kwargs:
            Additional keyword arguments that are passed to the
            `read_nc` function.
        """
        self._assert_read_mode()
        self._initialize_maps(skip_read=True)
        ncs = self.read_nc(fn, **kwargs)
        for name, ds in ncs.items():
            self.set_maps(ds, name=name)

    def write_maps(self, fn="maps/{name}.nc", **kwargs) -> None:
        r"""Write maps to netcdf file at <root>/<fn>.

        key-word arguments are passed to :py:meth:`~hydromt.models.Model.write_nc`

        Parameters
        ----------
        fn : str, optional
            filename relative to model root and should contain a {name} placeholder,
            by default 'maps/{name}.nc'
        \**kwargs:
            Additional keyword arguments that are passed to the
            `write_nc` function.
        """
        if len(self.maps) == 0:
            self.logger.debug("No maps data found, skip writing.")
        else:
            self._assert_write_mode()
            self.write_nc(self.maps, fn, **kwargs)

    # model geometry files
    @property
    def geoms(self) -> Dict[str, Union[gpd.GeoDataFrame, gpd.GeoSeries]]:
        """Model geometries.

        Return dict of geopandas.GeoDataFrame or geopandas.GeoDataSeries
        ..NOTE: previously call staticgeoms.
        """
        if self._geoms is None:
            self._initialize_geoms()
        return self._geoms

    def _initialize_geoms(self, skip_read=False) -> None:
        """Initialize geoms."""
        if self._geoms is None:
            self._geoms = dict()
            if self._read and not skip_read:
                self.read_geoms()

    def set_geoms(self, geom: Union[gpd.GeoDataFrame, gpd.GeoSeries], name: str):
        """Add data to the geoms attribute.

        Arguments
        ---------
        geom: geopandas.GeoDataFrame or geopandas.GeoSeries
            New geometry data to add
        name: str
            Geometry name.
        """
        self._initialize_geoms()
        gtypes = [gpd.GeoDataFrame, gpd.GeoSeries]
        if not np.any([isinstance(geom, t) for t in gtypes]):
            raise ValueError(
                "First parameter map(s) should be geopandas.GeoDataFrame"
                " or geopandas.GeoSeries"
            )
        if name in self._geoms:
            self.logger.warning(f"Replacing geom: {name}")
        if hasattr(self, "crs"):
            # Verify if a geom is set to model crs and if not sets geom to model crs
            if self.crs and self.crs != geom.crs:
                geom.to_crs(self.crs.to_epsg(), inplace=True)
        self._geoms[name] = geom

    def read_geoms(self, fn: str = "geoms/*.geojson", **kwargs) -> None:
        r"""Read model geometries files at <root>/<fn> and add to geoms property.

        key-word arguments are passed to :py:func:`geopandas.read_file`

        Parameters
        ----------
        fn : str, optional
            filename relative to model root, may contain wildcards,
            by default ``geoms/\*.nc``
        **kwargs:
            Additional keyword arguments that are passed to the
            `geopandas.read_file` function.
        """
        self._assert_read_mode()
        self._initialize_geoms(skip_read=True)
        fns = glob.glob(join(self.root, fn))
        for fn in fns:
            name = basename(fn).split(".")[0]
            self.logger.debug(f"Reading model file {name}.")
            self.set_geoms(gpd.read_file(fn, **kwargs), name=name)

    def write_geoms(
        self, fn: str = "geoms/{name}.geojson", to_wgs84: bool = False, **kwargs
    ) -> None:
        r"""Write model geometries to a vector file (by default GeoJSON) at <root>/<fn>.

        key-word arguments are passed to :py:meth:`geopandas.GeoDataFrame.to_file`

        Parameters
        ----------
        fn : str, optional
            filename relative to model root and should contain a {name} placeholder,
            by default 'geoms/{name}.geojson'
        to_wgs84: bool, optional
            Option to enforce writing GeoJSONs with WGS84(EPSG:4326) coordinates.
        \**kwargs:
            Additional keyword arguments that are passed to the
            `geopandas.to_file` function.
        """
        if len(self.geoms) == 0:
            self.logger.debug("No geoms data found, skip writing.")
            return
        self._assert_write_mode()
        for name, gdf in self.geoms.items():
            if not isinstance(gdf, (gpd.GeoDataFrame, gpd.GeoSeries)) or len(gdf) == 0:
                self.logger.warning(
                    f"{name} object of type {type(gdf).__name__} not recognized"
                )
                continue
            self.logger.debug(f"Writing file {fn.format(name=name)}")
            _fn = join(self.root, fn.format(name=name))
            if not isdir(dirname(_fn)):
                os.makedirs(dirname(_fn))
            if to_wgs84 and (
                kwargs.get("driver") == "GeoJSON"
                or str(fn).lower().endswith(".geojson")
            ):
                gdf = gdf.to_crs(4326)
            gdf.to_file(_fn, **kwargs)

    @property
    def staticgeoms(self):
        """Access the geometryes.

        This property will be deprecated in future versions,
        use :py:meth:`~hydromt.Model.geom`.
        """
        warnings.warn(
            "The staticgeoms method will be deprecated in future versions,"
            " use geoms instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        if self._geoms is None and self._read:
            self.read_staticgeoms()
        self._staticgeoms = self._geoms
        return self._staticgeoms

    def set_staticgeoms(self, geom: Union[gpd.GeoDataFrame, gpd.GeoSeries], name: str):
        """Set the geometries.

        This method will be deprecated in future versions,
        use :py:meth:`~hydromt.Model.set_geoms`.
        """
        warnings.warn(
            "The set_staticgeoms method will be deprecated in future versions,"
            " use set_geoms instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.set_geoms(geom, name)

    def read_staticgeoms(self):
        """Read gemoetries from disk.

        This method will be deprecated in future versions
        use :py:meth:`~hydromt.Model.read_geoms`.
        """
        warnings.warn(
            'The read_staticgeoms" method will be deprecated in future versions," \
            " use read_geoms instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        return self.read_geoms(fn="staticgeoms/*.geojson")

    def write_staticgeoms(self):
        """Write the geometries to disk.

        This method will be deprecated in future versions,
        use :py:meth:`~hydromt.Model.write_geoms`.
        """
        warnings.warn(
            'The "write_staticgeoms" method will be deprecated in future versions,"\
             " use  "write_geoms" instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        return self.write_geoms(fn="staticgeoms/{name}.geojson")

    # model forcing files
    @property
    def forcing(self) -> Dict[str, Union[xr.Dataset, xr.DataArray]]:
        """Model forcing. Returns dict of xarray.DataArray or xarray.Dataset."""
        if self._forcing is None:
            self._initialize_forcing()
        return self._forcing

    def _initialize_forcing(self, skip_read=False) -> None:
        """Initialize forcing."""
        if self._forcing is None:
            self._forcing = dict()
            if self._read and not skip_read:
                self.read_forcing()

    def set_forcing(
        self,
        data: Union[xr.DataArray, xr.Dataset, pd.DataFrame],
        name: Optional[str] = None,
        split_dataset: Optional[bool] = True,
    ):
        """Add data to forcing attribute.

        Data can be xarray.DataArray, xarray.Dataset or pandas.DataFrame.
        If pandas.DataFrame, indices should be the DataFrame index and the columns
        the variable names. the DataFrame will then be converted to xr.Dataset using
        :py:meth:`pandas.DataFrame.to_xarray` method.

        Arguments
        ---------
        data: xarray.Dataset or xarray.DataArray or pd.DataFrame
            New forcing data to add
        name: str, optional
            Results name, required if data is xarray.Dataset is and split_dataset=False.
        split_dataset: bool, optional
            If True (default), split a Dataset to store each variable as a DataArray.
        """
        self._initialize_forcing()
        if isinstance(data, pd.DataFrame):
            data = data.to_xarray()
        data_dict = _check_data(data, name, split_dataset)
        for name in data_dict:
            if name in self._forcing:
                self.logger.warning(f"Replacing forcing: {name}")
            self._forcing[name] = data_dict[name]

    def read_forcing(self, fn: str = "forcing/*.nc", **kwargs) -> None:
        """Read forcing at <root>/<fn> and add to forcing property.

        key-word arguments are passed to :py:meth:`~hydromt.models.Model.read_nc`

        Parameters
        ----------
        fn : str, optional
            filename relative to model root, may contain wildcards,
            by default forcing/.nc
        kwargs:
            Additional keyword arguments that are passed to the `read_nc`
            function.
        """
        self._assert_read_mode()
        self._initialize_forcing(skip_read=True)
        ncs = self.read_nc(fn, **kwargs)
        for name, ds in ncs.items():
            self.set_forcing(ds, name=name)

    def write_forcing(self, fn="forcing/{name}.nc", **kwargs) -> None:
        """Write forcing to netcdf file at <root>/<fn>.

        key-word arguments are passed to :py:meth:`~hydromt.models.Model.write_nc`

        Parameters
        ----------
        fn : str, optional
            filename relative to model root and should contain a {name} placeholder,
            by default 'forcing/{name}.nc'
        kwargs:
            Additional keyword arguments that are passed to the `write_nc`
            function.
        """
        if len(self.forcing) == 0:
            self.logger.debug("No forcing data found, skip writing.")
        else:
            self._assert_write_mode()
            self.write_nc(self.forcing, fn, **kwargs)

    # model state files
    @property
    def states(self) -> Dict[str, Union[xr.Dataset, xr.DataArray]]:
        """Model states. Returns dict of xarray.DataArray or xarray.Dataset."""
        if self._states is None:
            self._initialize_states()
        return self._states

    def _initialize_states(self, skip_read=False) -> None:
        """Initialize states."""
        if self._states is None:
            self._states = dict()
            if self._read and not skip_read:
                self.read_states()

    def set_states(
        self,
        data: Union[xr.DataArray, xr.Dataset],
        name: Optional[str] = None,
        split_dataset: Optional[bool] = True,
    ):
        """Add data to states attribute.

        Arguments
        ---------
        data: xarray.Dataset or xarray.DataArray
            New forcing data to add
        name: str, optional
            Results name, required if data is xarray.Dataset and split_dataset=False.
        split_dataset: bool, optional
            If True (default), split a Dataset to store each variable as a DataArray.
        """
        self._initialize_states()
        data_dict = _check_data(data, name, split_dataset)
        for name in data_dict:
            if name in self._states:
                self.logger.warning(f"Replacing state: {name}")
            self._states[name] = data_dict[name]

    def read_states(self, fn: str = "states/*.nc", **kwargs) -> None:
        r"""Read states at <root>/<fn> and add to states property.

        key-word arguments are passed to :py:meth:`~hydromt.models.Model.read_nc`

        Parameters
        ----------
        fn : str, optional
            filename relative to model root, may contain wildcards,
            by default states/\*.nc
        kwargs:
            Additional keyword arguments that are passed to the `read_nc`
            function.
        """
        self._assert_read_mode()
        self._initialize_states(skip_read=True)
        ncs = self.read_nc(fn, **kwargs)
        for name, ds in ncs.items():
            self.set_states(ds, name=name, split_dataset=True)

    def write_states(self, fn="states/{name}.nc", **kwargs) -> None:
        """Write states to netcdf file at <root>/<fn>.

        key-word arguments are passed to :py:meth:`~hydromt.models.Model.write_nc`

        Parameters
        ----------
        fn : str, optional
            filename relative to model root and should contain a {name} placeholder,
            by default 'states/{name}.nc'
        **kwargs:
            Additional keyword arguments that are passed to the `write_nc`
            function.
        """
        if len(self.states) == 0:
            self.logger.debug("No states data found, skip writing.")
        else:
            self._assert_write_mode()
            self.write_nc(self.states, fn, **kwargs)

    # model results files; NOTE we don't have a write_results method
    # (that's up to the model kernel)
    @property
    def results(self) -> Dict[str, Union[xr.Dataset, xr.DataArray]]:
        """Model results. Returns dict of xarray.DataArray or xarray.Dataset."""
        if self._results is None:
            self._initialize_results()
        return self._results

    def _initialize_results(self, skip_read=False) -> None:
        """Initialize results."""
        if self._results is None:
            self._results = dict()
            if self._read and not skip_read:
                self.read_results()

    def set_results(
        self,
        data: Union[xr.DataArray, xr.Dataset],
        name: Optional[str] = None,
        split_dataset: Optional[bool] = False,
    ):
        """Add data to results attribute.

        Dataset can either be added as is (default) or split into several
        DataArrays using the split_dataset argument.

        Arguments
        ---------
        data: xarray.Dataset or xarray.DataArray
            New forcing data to add
        name: str, optional
            Results name, required if data is xarray.Dataset and split_dataset=False.
        split_dataset: bool, optional
            If True (False by default), split a Dataset to store each variable
            as a DataArray.
        """
        self._initialize_results()
        data_dict = _check_data(data, name, split_dataset)
        for name in data_dict:
            if name in self._results:
                self.logger.warning(f"Replacing result: {name}")
            self._results[name] = data_dict[name]

    def read_results(self, fn: str = "results/*.nc", **kwargs) -> None:
        """Read results at <root>/<fn> and add to results property.

        key-word arguments are passed to :py:meth:`~hydromt.models.Model.read_nc`

        Parameters
        ----------
        fn : str, optional
            filename relative to model root, may contain wildcards,
            by default ``results/*.nc``
        kwargs:
            Additional keyword arguments that are passed to the `read_nc`
            function.
        """
        self._assert_read_mode()
        self._initialize_results(skip_read=True)
        ncs = self.read_nc(fn, **kwargs)
        for name, ds in ncs.items():
            self.set_results(ds, name=name)

    # general reader & writer
    def _cleanup(self, forceful_overwrite=False, max_close_attempts=2) -> List[str]:
        """Try to close all defered file handles.

        Try to overwrite the destination file with the temporary one until either the
        maximum number of tries is reached or until it succeeds. The forced cleanup
        also attempts to close the original file handle, which could cause trouble
        if the user will try to read from the same file handle after this function
        is called.

        Parameters
        ----------
        forceful_overwrite: bool
            Attempt to force closing defered file handles before writing to them.
        max_close_attempts: int
            Number of times to try and overwrite the original file, before giving up.

        """
        failed_closes = []
        while len(self._defered_file_closes) > 0:
            close_handle = self._defered_file_closes.pop()
            if close_handle["close_attempts"] > max_close_attempts:
                # already tried to close this to many times so give up
                self.logger.error(
                    f"Max write attempts to file {close_handle['org_fn']}"
                    " exceeded. Skipping..."
                    f"Instead data was written to tmpfile: {close_handle['tmp_fn']}"
                )
                continue

            if forceful_overwrite:
                close_handle["ds"].close()
            try:
                shutil.move(close_handle["tmp_fn"], close_handle["org_fn"])
            except PermissionError:
                self.logger.error(
                    f"Could not write to destination file {close_handle['org_fn']} "
                    "because the following error was raised: {e}"
                )
                close_handle["close_attempts"] += 1
                self._defered_file_closes.append(close_handle)
                failed_closes.append((close_handle["org_fn"], close_handle["tmp_fn"]))

        return list(set(failed_closes))

    def write_nc(
        self,
        nc_dict: XArrayDict,
        fn: str,
        gdal_compliant: bool = False,
        rename_dims: bool = False,
        force_sn: bool = False,
        **kwargs,
    ) -> None:
        """Write dictionnary of xarray.Dataset and/or xarray.DataArray to netcdf files.

        Possibility to update the xarray objects attributes to get GDAL compliant NetCDF
        files, using :py:meth:`~hydromt.raster.gdal_compliant`.
        The function will first try to directly write to file. In case of
        PermissionError, it will first write a temporary file and add to the
        self._defered_file_closes attribute. Renaming and closing of netcdf filehandles
        will be done by calling the self._cleanup function.

        key-word arguments are passed to :py:meth:`xarray.Dataset.to_netcdf`

        Parameters
        ----------
        nc_dict: dict
            Dictionary of xarray.Dataset and/or xarray.DataArray to write
        fn: str
            filename relative to model root and should contain a {name} placeholder
        gdal_compliant: bool, optional
            If True, convert xarray.Dataset and/or xarray.DataArray to gdal compliant
            format using :py:meth:`~hydromt.raster.gdal_compliant`
        rename_dims: bool, optional
            If True, rename x_dim and y_dim to standard names depending on the CRS
            (x/y for projected and lat/lon for geographic). Only used if
            ``gdal_compliant`` is set to True. By default, False.
        force_sn: bool, optional
            If True, forces the dataset to have South -> North orientation. Only used
            if ``gdal_compliant`` is set to True. By default, False.
        **kwargs:
            Additional keyword arguments that are passed to the `to_netcdf`
            function.
        """
        for name, ds in nc_dict.items():
            if not isinstance(ds, (xr.Dataset, xr.DataArray)) or len(ds) == 0:
                self.logger.error(
                    f"{name} object of type {type(ds).__name__} not recognized"
                )
                continue
            self.logger.debug(f"Writing file {fn.format(name=name)}")
            _fn = join(self.root, fn.format(name=name))
            if not isdir(dirname(_fn)):
                os.makedirs(dirname(_fn))
            if gdal_compliant:
                ds = ds.raster.gdal_compliant(
                    rename_dims=rename_dims, force_sn=force_sn
                )
            try:
                ds.to_netcdf(_fn, **kwargs)
            except PermissionError:
                logger.warning(f"Could not write to file {_fn}, defering write")
                if self._TMP_DATA_DIR is None:
                    self._TMP_DATA_DIR = TemporaryDirectory()

                tmp_fn = join(str(self._TMP_DATA_DIR), f"{_fn}.tmp")
                ds.to_netcdf(tmp_fn, **kwargs)
                self._defered_file_closes.append(
                    DeferedFileClose(
                        ds=ds,
                        org_fn=join(str(self._TMP_DATA_DIR), _fn),
                        tmp_fn=tmp_fn,
                        close_attempts=1,
                    )
                )

    def read_nc(
        self,
        fn: StrPath,
        mask_and_scale: bool = False,
        single_var_as_array: bool = True,
        load: bool = False,
        **kwargs,
    ) -> Dict[str, xr.Dataset]:
        """Read netcdf files at <root>/<fn> and return as dict of xarray.Dataset.

        NOTE: Unless `single_var_as_array` is set to False a single-variable data source
        will be returned as :py:class:`xarray.DataArray` rather than
        :py:class:`xarray.Dataset`.
        key-word arguments are passed to :py:func:`xarray.open_dataset`.

        Parameters
        ----------
        fn : str
            filename relative to model root, may contain wildcards
        mask_and_scale : bool, optional
            If True, replace array values equal to _FillValue with NA and scale values
            according to the formula original_values * scale_factor + add_offset, where
            _FillValue, scale_factor and add_offset are taken from variable attributes
            (if they exist).
        single_var_as_array : bool, optional
            If True, return a DataArray if the dataset consists of a single variable.
            If False, always return a Dataset. By default True.
        load : bool, optional
            If True, the data is loaded into memory. By default False.
        **kwargs:
            Additional keyword arguments that are passed to the `xr.open_dataset`
            function.

        Returns
        -------
        Dict[str, xr.Dataset]
            dict of xarray.Dataset
        """
        ncs = dict()
        fns = glob.glob(join(self.root, fn))
        if "chunks" not in kwargs:  # read lazy by default
            kwargs.update(chunks="auto")
        for fn in fns:
            name = basename(fn).split(".")[0]
            self.logger.debug(f"Reading model file {name}.")
            # Load data to allow overwritting in r+ mode
            if load:
                ds = xr.open_dataset(fn, mask_and_scale=mask_and_scale, **kwargs).load()
                ds.close()
            else:
                ds = xr.open_dataset(fn, mask_and_scale=mask_and_scale, **kwargs)
            # set geo coord if present as coordinate of dataset
            if GEO_MAP_COORD in ds.data_vars:
                ds = ds.set_coords(GEO_MAP_COORD)
            # single-variable Dataset to DataArray
            if single_var_as_array and len(ds.data_vars) == 1:
                (ds,) = ds.data_vars.values()
            ncs.update({name: ds})
        return ncs

    ## properties / methods below can be used directly in actual class
    @property
    def crs(self) -> CRS:
        """Returns coordinate reference system embedded in region."""
        if len(self.staticmaps) > 0:
            return self.staticmaps.raster.crs
        else:
            return self.region.crs

    def set_crs(self, crs) -> None:
        """Set the coordinate reference system."""
        warnings.warn(
            '"set_crs" is deprecated. Please set the crs of all model'
            " components instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        if len(self.staticmaps) > 0:
            return self.staticmaps.raster.set_crs(crs)

    @property
    def dims(self) -> Tuple:
        """Returns spatial dimension names of staticmaps.

        ..NOTE: will be deprecated in future versions.
        """
        if len(self.staticmaps) > 0:
            return self.staticmaps.raster.dims

    @property
    def coords(self) -> Dict:
        """Returns the coordinates of model staticmaps.

        ..NOTE: will be deprecated in future versions.
        """
        if len(self.staticmaps) > 0:
            return self.staticmaps.raster.coords

    @property
    def res(self) -> Tuple:
        """Returns the resolution of the model staticmaps.

        ..NOTE: will be deprecated in future versions.
        """
        if len(self.staticmaps) > 0:
            return self.staticmaps.raster.res

    @property
    def transform(self):
        """Returns the geospatial transform of the model staticmaps.

        ..NOTE: will be deprecated in future versions.
        """
        if len(self.staticmaps) > 0:
            return self.staticmaps.raster.transform

    @property
    def width(self):
        """Returns the width of the model staticmaps.

        ..NOTE: will be deprecated in future versions.
        """
        if len(self.staticmaps) > 0:
            return self.staticmaps.raster.width

    @property
    def height(self):
        """Returns the height of the model staticmaps.

        ..NOTE: will be deprecated in future versions.
        """
        if len(self.staticmaps) > 0:
            return self.staticmaps.raster.height

    @property
    def shape(self) -> Tuple:
        """Returns the shape of the model staticmaps.

        ..NOTE: will be deprecated in future versions.
        """
        if len(self.staticmaps) > 0:
            return self.staticmaps.raster.shape

    @property
    def bounds(self) -> Tuple:
        """Returns the bounding box of the model region.

        ..NOTE: will be deprecated in future versions.
        """
        if len(self.staticmaps) > 0:
            return self.staticmaps.raster.bounds
        else:
            return self.region.total_bounds

    @property
    def region(self) -> gpd.GeoDataFrame:
        """Returns the geometry of the model area of interest."""
        region = gpd.GeoDataFrame()
        if "region" in self.geoms:
            region = self.geoms["region"]
        # TODO: For now stays here but move to grid in GridModel and delete
        elif len(self.staticmaps) > 0:
            warnings.warn(
                'Defining "region" based on staticmaps will be deprecated. Either use'
                " region from GridModel or define your own method.",
                DeprecationWarning,
                stacklevel=2,
            )
            crs = self.staticmaps.raster.crs
            if crs is None and hasattr(crs, "to_epsg"):
                crs = crs.to_epsg()  # not all CRS have an EPSG code
            region = gpd.GeoDataFrame(
                geometry=[box(*self.staticmaps.raster.bounds)], crs=crs
            )
        return region

    # test methods
    def test_model_api(self):
        """Test compliance with HydroMT Model API."""
        warnings.warn(
            '"test_model_api" is now part of the internal API, use "_test_model_api"'
            " instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._test_model_api()

    def _test_model_api(self) -> List:
        """Test compliance with HydroMT Model API.

        Returns
        -------
        non_compliant: list
            List of model components that are non-compliant with the model API
            structure.
        """
        non_compliant = []
        for component, dtype in self.api.items():
            obj = getattr(self, component, None)
            try:
                assert obj is not None, component
                _assert_isinstance(obj, dtype, component)
            except AssertionError as err:
                non_compliant.append(str(err))

        return non_compliant

    def _test_equal(self, other, skip_component=None) -> Tuple[bool, Dict]:
        """Test if two models including their data components are equal.

        Parameters
        ----------
        other : Model (or subclass)
            Model to compare against
        skip_component: list
            List of components to skip when testing equality. By default root.

        Returns
        -------
        equal: bool
            True if equal
        errors: dict
            Dictionary with errors per model component which is not equal
        """
        skip_component = skip_component or ["root"]
        assert isinstance(other, type(self))
        components = list(self.api.keys())
        components_other = list(other.api.keys())
        assert components == components_other
        for cp in skip_component:
            if cp in components:
                components.remove(cp)
        errors = {}
        for prop in components:
            errors.update(
                **_check_equal(getattr(self, prop), getattr(other, prop), prop)
            )
        return len(errors) == 0, errors


def _check_data(
    data: Union[xr.DataArray, xr.Dataset],
    name: Optional[str] = None,
    split_dataset=True,
) -> Dict:
    if isinstance(data, xr.DataArray):
        # NOTE name can be different from data.name !
        if data.name is None and name is not None:
            data.name = name
        elif name is None and data.name is not None:
            name = data.name
        elif data.name is None and name is None:
            raise ValueError("Name required for DataArray.")
        data = {name: data}
    elif isinstance(data, xr.Dataset):  # return dict for consistency
        if split_dataset:
            data = {name: data[name] for name in data.data_vars}
        elif name is None:
            raise ValueError("Name required for Dataset.")
        else:
            data = {name: data}
    else:
        raise ValueError(f'Data type "{type(data).__name__}" not recognized')
    return data


def _assert_isinstance(obj: Any, dtype: Any, name: str = ""):
    """Check if obj match typing or class (dtype)."""
    args = typing.get_args(dtype)
    _cls = typing.get_origin(dtype)
    if len(args) == 0 and dtype != Any and dtype is not None:
        assert isinstance(obj, dtype), name
    elif _cls == Union:
        assert isinstance(obj, args), name
    elif _cls is not None:
        assert isinstance(obj, _cls), name
    # recursive check of dtype dict keys and values
    if len(args) > 0 and _cls is dict:
        for key, val in obj.items():
            _assert_isinstance(key, args[0], f"{name}.{str(key)}")
            _assert_isinstance(val, args[1], f"{name}.{str(key)}")


def _check_equal(a, b, name="") -> Dict[str, str]:
    """Recursive test of model components.

    Returns dict with component name and associated error message.
    """
    errors = {}
    try:
        assert isinstance(b, type(a)), "property types do not match"
        if isinstance(a, dict):
            for key in a:
                assert key in b, f"{key} missing"
                errors.update(**_check_equal(a[key], b[key], f"{name}.{key}"))
        elif isinstance(a, (xr.DataArray, xr.Dataset)):
            xr.testing.assert_allclose(a, b)
        elif isinstance(a, gpd.GeoDataFrame):
            assert_geodataframe_equal(a, b, check_like=True, check_less_precise=True)
        elif isinstance(a, np.ndarray):
            np.testing.assert_allclose(a, b)
        else:
            assert a == b, "values not equal"
    except AssertionError as e:
        errors.update({name: e})
    return errors
